# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import six
import contextlib
import json

from datetime import datetime, timedelta

import flask

from werkzeug.exceptions import BadRequest
from freezegun import freeze_time
from mock import patch, PropertyMock, call

import odcs.server.auth

from odcs.server import conf, db, app, login_manager, version
from odcs.server.models import Compose, User
from odcs.common.types import (
    COMPOSE_STATES,
    COMPOSE_RESULTS,
    COMPOSE_FLAGS,
    MULTILIB_METHODS,
)
from odcs.server.pungi import PungiSourceType
from .utils import ModelsBaseTest
from odcs.server.api_utils import validate_json_data

import unittest


@login_manager.user_loader
def user_loader(username):
    return User.find_user_by_name(username=username)


class TestValidateJSONData(unittest.TestCase):
    def test_validate_json_data_allowed_dict(self):
        data = {"source": {"source": ""}}
        validate_json_data(data)

    def test_validate_json_data_unallowed_dict(self):
        data = {"source": {"source": {"k": "v"}}}
        self.assertRaises(ValueError, validate_json_data, data)

    def test_validate_json_data_unallowed_dict_toplevel(self):
        data = {"source2": {"source": {"k": "v"}}}
        self.assertRaises(ValueError, validate_json_data, data)

    def test_validate_json_data_list(self):
        data = {"source": ["1", "2"]}
        validate_json_data(data)

    def test_validate_json_data_list_unallowed_chars(self):
        for char in ["\n", "{", "}", "'", '"']:
            data = {"source": ["%s1" % char, "2"]}
            self.assertRaises(ValueError, validate_json_data, data)

    def test_validate_json_data_str_unallowed_chars(self):
        for char in ["\n", "{", "}", "'", '"']:
            data = {"source": {"x": char + "1"}}
            self.assertRaises(ValueError, validate_json_data, data)

    def test_validate_json_data_int(self):
        data = {"source": {"x": 1}}
        validate_json_data(data)

    def test_validate_json_data_float(self):
        data = {"source": {"x": 1.5}}
        validate_json_data(data)

    def test_validate_json_data_unallowed_type(self):
        data = {"source": {"x": PropertyMock()}}
        self.assertRaises(ValueError, validate_json_data, data)

    def test_validate_json_data_allowed_package(self):
        data = {"packages": ["gcc-g++"]}
        validate_json_data(data)


class ViewBaseTest(ModelsBaseTest):
    def setUp(self):
        super(ViewBaseTest, self).setUp()

        patched_allowed_clients = {
            "groups": {
                "composer": {},
                "dev2": {"source_types": ["module"]},
                "dev3": {"source_types": ["raw_config"]},
            },
            "users": {
                "dev": {"arches": ["ppc64", "s390", "x86_64"]},
                "dev2": {
                    "source_types": ["module", "raw_config"],
                    "compose_types": ["test", "nightly"],
                },
                "dev3": {"source_types": ["tag"], "target_dirs": ["releng-private"]},
                "dev4": {
                    "source_types": ["raw_config"],
                    "raw_config_keys": ["pungi_cfg2"],
                },
            },
        }
        patched_admins = {"groups": ["admin"], "users": ["root"]}
        self.patch_allowed_clients = patch.object(
            odcs.server.auth.conf, "allowed_clients", new=patched_allowed_clients
        )
        self.patch_admins = patch.object(
            odcs.server.auth.conf, "admins", new=patched_admins
        )
        self.patch_allowed_clients.start()
        self.patch_admins.start()

        self.client = app.test_client()

        self.setup_test_data()

    def tearDown(self):
        super(ViewBaseTest, self).tearDown()

        self.patch_allowed_clients.stop()
        self.patch_admins.stop()

    @contextlib.contextmanager
    def test_request_context(self, user=None, groups=None, **kwargs):
        with app.test_request_context(**kwargs):
            patch_auth_backend = None
            if user is not None:
                # authentication is disabled with auth_backend=noauth
                patch_auth_backend = patch.object(
                    odcs.server.auth.conf, "auth_backend", new="kerberos"
                )
                patch_auth_backend.start()
                if not User.find_user_by_name(user):
                    User.create_user(username=user)
                    db.session.commit()
                flask.g.user = User.find_user_by_name(user)

                if groups is not None:
                    if isinstance(groups, list):
                        flask.g.groups = groups
                    else:
                        flask.g.groups = [groups]
                else:
                    flask.g.groups = []
                with self.client.session_transaction() as sess:
                    sess["user_id"] = user
                    sess["_fresh"] = True
            try:
                yield
            finally:
                if patch_auth_backend is not None:
                    patch_auth_backend.stop()

    def setup_test_data(self):
        """Set up data for running tests"""


class TestOpenIDCLogin(ViewBaseTest):
    """Test that OpenIDC login"""

    def setUp(self):
        super(TestOpenIDCLogin, self).setUp()
        self.patch_auth_backend = patch.object(
            odcs.server.auth.conf, "auth_backend", new="openidc"
        )
        self.patch_auth_backend.start()

    def tearDown(self):
        super(TestOpenIDCLogin, self).tearDown()
        self.patch_auth_backend.stop()

    def test_openidc_post_unauthorized(self):
        rv = self.client.post("/api/1/composes/", data="")
        self.assertEqual(rv.status, "401 UNAUTHORIZED")

    def test_openidc_patch_unauthorized(self):
        rv = self.client.patch("/api/1/composes/1")
        self.assertEqual(rv.status, "401 UNAUTHORIZED")

    def test_openidc_delete_unauthorized(self):
        rv = self.client.delete("/api/1/composes/1")
        self.assertEqual(rv.status, "401 UNAUTHORIZED")


class TestHandlingErrors(ViewBaseTest):
    """Test registered error handlers"""

    @patch("odcs.server.views.ODCSAPI.delete")
    def test_bad_request_error(self, delete):
        delete.side_effect = BadRequest("bad request to delete")

        resp = self.client.delete("/api/1/composes/100")
        data = json.loads(resp.get_data(as_text=True))

        self.assertEqual("Bad Request", data["error"])
        self.assertEqual(400, data["status"])
        self.assertIn("bad request to delete", data["message"])

    def test_return_internal_server_error_if_error_is_not_caught(self):
        possible_errors = [
            RuntimeError("runtime error"),
            IndexError("out of scope"),
            OSError("os error"),
        ]
        for e in possible_errors:
            with patch("odcs.server.views.filter_composes", side_effect=e):
                resp = self.client.get("/api/1/composes/")
                data = json.loads(resp.get_data(as_text=True))

                self.assertEqual("Internal Server Error", data["error"])
                self.assertEqual(500, data["status"])
                self.assertEqual(str(e), data["message"])


class TestViews(ViewBaseTest):
    maxDiff = None

    def setUp(self):
        super(TestViews, self).setUp()
        self.oidc_base_namespace = patch.object(
            conf, "oidc_base_namespace", new="http://example.com/"
        )
        self.oidc_base_namespace.start()

    def tearDown(self):
        self.oidc_base_namespace.stop()
        super(TestViews, self).tearDown()

    def setup_test_data(self):
        self.initial_datetime = datetime(
            year=2016, month=1, day=1, hour=0, minute=0, second=0
        )
        with freeze_time(self.initial_datetime):
            self.c1 = Compose.create(
                db.session,
                "unknown",
                PungiSourceType.MODULE,
                "testmodule:master",
                COMPOSE_RESULTS["repository"],
                60,
            )
            self.c2 = Compose.create(
                db.session,
                "me",
                PungiSourceType.KOJI_TAG,
                "f26",
                COMPOSE_RESULTS["repository"],
                60,
            )
            db.session.add(self.c1)
            db.session.add(self.c2)
            db.session.commit()

    def test_metrics(self):
        rv = self.client.get("/api/1/metrics/")
        data = rv.get_data(as_text=True)
        self.assertTrue("HELP composes_total Total number of composes" in data)

    def test_index(self):
        rv = self.client.get("/")
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv.content_type, "text/html; charset=utf-8")
        data = rv.get_data(as_text=True)
        self.assertIn("On Demand Compose Service", data)

    def test_about(self):
        rv = self.client.get("/api/1/about/")
        data = json.loads(rv.get_data(as_text=True))
        self.assertEqual(
            data,
            {
                "version": version,
                "auth_backend": "noauth",
                "raw_config_urls": {},
                "allowed_clients": odcs.server.auth.conf.allowed_clients,
                "sigkeys": [],
            },
        )

    def test_submit_invalid_json(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post("/api/1/composes/", data="{")
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(rv.status, "400 BAD REQUEST")
        self.assertEqual(data["error"], "Bad Request")
        self.assertEqual(data["status"], 400)
        self.assertTrue(data["message"].find("Failed to decode JSON object") != -1)

    def test_submit_build(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "module", "source": "testmodule:master"}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        expected_json = {
            "source_type": 2,
            "state": 0,
            "time_done": None,
            "state_name": "wait",
            "state_reason": None,
            "source": u"testmodule:master",
            "owner": u"dev",
            "result_repo": "http://localhost/odcs/odcs-%d/compose/Temporary"
            % data["id"],
            "result_repofile": "http://localhost/odcs/odcs-%d/compose/Temporary/odcs-%d.repo"
            % (data["id"], data["id"]),
            "time_submitted": data["time_submitted"],
            "id": data["id"],
            "time_started": None,
            "time_removed": None,
            "removed_by": None,
            "time_to_expire": data["time_to_expire"],
            "flags": [],
            "results": ["repository"],
            "sigkeys": "",
            "koji_event": None,
            "koji_task_id": None,
            "packages": None,
            "builds": None,
            "arches": "x86_64",
            "multilib_arches": "",
            "multilib_method": 0,
            "lookaside_repos": "",
            "modular_koji_tags": None,
            "module_defaults_url": None,
            "label": None,
            "compose_type": "test",
            "pungi_compose_id": None,
            "target_dir": "default",
            "scratch_modules": None,
            "toplevel_url": "http://localhost/odcs/odcs-%d" % data["id"],
            "parent_pungi_compose_ids": None,
            "scratch_build_tasks": None,
        }
        self.assertEqual(data, expected_json)

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

    def test_submit_build_no_packages(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "tag", "source": "f26"}, "flags": ["no_deps"]}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["state_name"], "wait")

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])
        self.assertEqual(c.packages, None)

    def test_submit_build_nodeps(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "tag", "source": "f26", "packages": ["ed"]},
                        "flags": ["no_deps"],
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["flags"], ["no_deps"])

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 3).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])
        self.assertEqual(c.flags, COMPOSE_FLAGS["no_deps"])

    def test_submit_build_noinheritance(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "tag", "source": "f26", "packages": ["ed"]},
                        "flags": ["no_inheritance"],
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["flags"], ["no_inheritance"])

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 3).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])
        self.assertEqual(c.flags, COMPOSE_FLAGS["no_inheritance"])

    def test_submit_build_boot_iso(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "tag", "source": "f26", "packages": ["ed"]},
                        "results": ["boot.iso"],
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(set(data["results"]), set(["repository", "boot.iso"]))

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 3).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])
        self.assertEqual(
            c.results, COMPOSE_RESULTS["boot.iso"] | COMPOSE_RESULTS["repository"]
        )

    def test_submit_build_with_koji_event(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {
                            "type": "tag",
                            "source": "f26",
                            "packages": ["ed"],
                            "koji_event": 123456,
                        }
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["koji_event"], 123456)

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 3).one()
        self.assertEqual(c.koji_event, 123456)

    def test_submit_build_sigkeys(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {
                            "type": "tag",
                            "source": "f26",
                            "packages": ["ed"],
                            "sigkeys": ["123", "456"],
                        }
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["sigkeys"], "123 456")

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

    @patch.object(odcs.server.config.Config, "sigkeys", new_callable=PropertyMock)
    def test_submit_build_default_sigkeys(self, sigkeys):
        with self.test_request_context(user="dev"):
            sigkeys.return_value = ["x", "y"]
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "tag", "source": "f26", "packages": ["ed"]}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["sigkeys"], "x y")

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

    def test_submit_build_arches(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "tag", "source": "f26", "packages": ["ed"]},
                        "arches": ["ppc64", "s390"],
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["arches"], "ppc64 s390")

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

    def test_submit_build_multilib_arches(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "tag", "source": "f26", "packages": ["ed"]},
                        "arches": ["ppc64", "s390"],
                        "multilib_arches": ["x86_64", "ppc64le"],
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["multilib_arches"], "x86_64 ppc64le")

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

    def test_submit_build_multilib_method(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "tag", "source": "f26", "packages": ["ed"]},
                        "arches": ["ppc64", "s390"],
                        "multilib_method": ["runtime", "devel"],
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(
            data["multilib_method"],
            MULTILIB_METHODS["runtime"] | MULTILIB_METHODS["devel"],
        )

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

    def test_submit_build_multilib_method_unknown(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "tag", "source": "f26", "packages": ["ed"]},
                        "arches": ["ppc64", "s390"],
                        "multilib_method": ["foo", "devel"],
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["message"], 'Unknown multilib method "foo"')

    def test_submit_build_modular_koji_tags(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {
                            "type": "tag",
                            "source": "f26",
                            "modular_koji_tags": ["f26-modules"],
                        }
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["modular_koji_tags"], "f26-modules")

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

    def test_submit_build_target_dir_unknown(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "tag", "source": "f26"}, "target_dir": "foo"}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["status"], 400)
        self.assertEqual(data["error"], "Bad Request")
        self.assertEqual(data["message"], 'Unknown "target_dir" "foo"')

    @patch.object(
        odcs.server.config.Config,
        "extra_target_dirs",
        new={"releng-private": "/tmp/private"},
    )
    def test_submit_build_target_not_allowed(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "tag", "source": "f26"},
                        "target_dir": "releng-private",
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["status"], 403)
        self.assertEqual(data["error"], "Forbidden")
        self.assertEqual(
            data["message"],
            "User dev not allowed to operate with compose with target_dirs=releng-private.",
        )

    @patch.object(
        odcs.server.config.Config,
        "extra_target_dirs",
        new={"releng-private": "/tmp/private"},
    )
    def test_submit_build_target_dir(self):
        with self.test_request_context(user="dev3"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "tag", "source": "f26"},
                        "target_dir": "releng-private",
                    }
                ),
            )
            self.assertEqual(rv.status, "200 OK")
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["target_dir"], "releng-private")

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 3).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])
        self.assertEqual(c.target_dir, "/tmp/private")

    def test_submit_build_module_defaults_url(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {
                            "type": "tag",
                            "source": "f26",
                            "module_defaults_url": "git://localhost.tld/x.git",
                            "module_defaults_commit": "master",
                        }
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(
            data["module_defaults_url"], "git://localhost.tld/x.git master"
        )

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

    def test_submit_build_module_defaults_url_no_branch(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {
                            "type": "tag",
                            "source": "f26",
                            "module_defaults_url": "git://localhost.tld/x.git",
                        }
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))
            self.assertEqual(data["status"], 400)
            self.assertEqual(data["error"], "Bad Request")
            self.assertEqual(
                data["message"],
                'The "module_defaults_url" and "module_defaults_commit" '
                "must be used together.",
            )

    def test_submit_build_duplicate_sources(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "module", "source": "foo:x foo:x foo:y"}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["source"].count("foo:x"), 1)
        self.assertEqual(data["source"].count("foo:y"), 1)

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

    def test_submit_build_extra_builds(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {
                            "type": "tag",
                            "source": "f26",
                            "packages": ["ed"],
                            "builds": ["foo-1-1", "bar-1-1"],
                        }
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["builds"], "foo-1-1 bar-1-1")

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 3).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])
        self.assertEqual(c.builds, "foo-1-1 bar-1-1")

    def test_submit_build_source_type_build(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {
                            "type": "build",
                            "packages": ["ed"],
                            "builds": ["foo-1-1", "bar-1-1"],
                        }
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["builds"], "foo-1-1 bar-1-1")

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 3).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])
        self.assertEqual(c.source_type, PungiSourceType.BUILD)
        self.assertEqual(c.builds, "foo-1-1 bar-1-1")

    def test_submit_build_resurrection_removed(self):
        self.c1.state = COMPOSE_STATES["removed"]
        self.c1.reused_id = 1
        db.session.commit()

        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "renew-compose")
            ]

            rv = self.client.patch("/api/1/composes/1")
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["id"], 3)
        self.assertEqual(data["state_name"], "wait")
        self.assertEqual(data["source"], "testmodule:master")
        self.assertEqual(data["time_removed"], None)

        c = db.session.query(Compose).filter(Compose.id == 3).one()
        self.assertEqual(c.reused_id, None)

    def test_submit_build_resurrection_not_allowed_raw_config_key(self):
        self.c1.state = COMPOSE_STATES["removed"]
        self.c1.reused_id = 1
        self.c1.source_type = 5
        self.c1.source = "pungi_cfg#hash"
        db.session.commit()

        with self.test_request_context(user="dev4"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "renew-compose")
            ]

            rv = self.client.patch("/api/1/composes/1")
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(
            data["message"],
            "User dev4 not allowed to operate with compose with raw_config_keys=pungi_cfg.",
        )

    def test_submit_build_resurrection_allowed_raw_config_key(self):
        self.c1.state = COMPOSE_STATES["removed"]
        self.c1.reused_id = 1
        self.c1.source_type = 5
        self.c1.source = "pungi_cfg2#hash"
        db.session.commit()

        with self.test_request_context(user="dev4"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "renew-compose")
            ]

            rv = self.client.patch("/api/1/composes/1")
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["id"], 3)
        self.assertEqual(data["state_name"], "wait")

    def test_submit_build_resurrection_failed(self):
        self.c1.state = COMPOSE_STATES["failed"]
        self.c1.reused_id = 1
        db.session.commit()

        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "renew-compose")
            ]

            rv = self.client.patch("/api/1/composes/1")
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["id"], 3)
        self.assertEqual(data["state_name"], "wait")
        self.assertEqual(data["source"], "testmodule:master")
        self.assertEqual(data["time_removed"], None)

        c = db.session.query(Compose).filter(Compose.id == 3).one()
        self.assertEqual(c.reused_id, None)

    def test_submit_build_resurrection_no_removed(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "renew-compose")
            ]

            rv = self.client.patch("/api/1/composes/1")
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["message"], "No compose with id 1 found")

    def test_submit_build_resurrection_not_found(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "renew-compose")
            ]

            rv = self.client.patch("/api/1/composes/100")
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["message"], "No compose with id 100 found")

    def test_submit_build_resurrection_removed_new_sigkeys(self):
        self.c1.state = COMPOSE_STATES["removed"]
        self.c1.reused_id = 1
        db.session.commit()

        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "renew-compose")
            ]

            rv = self.client.patch(
                "/api/1/composes/1", data=json.dumps({"sigkeys": ["123", "456"]})
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["id"], 3)
        self.assertEqual(data["state_name"], "wait")
        self.assertEqual(data["source"], "testmodule:master")
        self.assertEqual(data["sigkeys"], "123 456")
        self.assertEqual(data["time_removed"], None)

        c = db.session.query(Compose).filter(Compose.id == 3).one()
        self.assertEqual(c.reused_id, None)

    def test_submit_build_resurrection_done_new_sigkeys(self):
        self.c1.state = COMPOSE_STATES["done"]
        self.c1.reused_id = 1
        self.c1.sigkeys = "012 345"
        db.session.commit()

        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "renew-compose")
            ]

            rv = self.client.patch(
                "/api/1/composes/1", data=json.dumps({"sigkeys": ["123", "456"]})
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["id"], 3)
        self.assertEqual(data["state_name"], "wait")
        self.assertEqual(data["source"], "testmodule:master")
        self.assertEqual(data["sigkeys"], "123 456")
        self.assertEqual(data["time_removed"], None)

        c = db.session.query(Compose).filter(Compose.id == 3).one()
        self.assertEqual(c.reused_id, None)

    def test_submit_build_not_allowed_source_type(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps({"source": {"type": "repo", "source": "/path"}}),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(
            data["message"],
            "User dev not allowed to operate with compose with source_types=repo.",
        )

    @patch.object(
        odcs.server.config.Config,
        "raw_config_urls",
        new={"pungi_cfg": "http://localhost/pungi.conf#%s"},
    )
    def test_submit_build_not_allowed_compose_type(self):
        with self.test_request_context(user="dev2"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "raw_config", "source": "pungi_cfg#hash"},
                        "compose_type": "production",
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(
            data["message"],
            "User dev2 not allowed to operate with compose with compose_types=production.",
        )

    @patch.object(
        odcs.server.config.Config,
        "raw_config_urls",
        new={"pungi_cfg": "http://localhost/pungi.conf#%s"},
    )
    def test_submit_build_not_allowed_raw_config_key(self):
        with self.test_request_context(user="dev4"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "raw_config", "source": "pungi_cfg#hash"},
                        "compose_type": "production",
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(
            data["message"],
            "User dev4 not allowed to operate with compose with raw_config_keys=pungi_cfg.",
        )

    @patch.object(
        odcs.server.config.Config,
        "raw_config_urls",
        new={"pungi_cfg2": "http://localhost/pungi.conf#%s"},
    )
    def test_submit_build_allowed_raw_config_key(self):
        with self.test_request_context(user="dev4"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "raw_config", "source": "pungi_cfg2#hash"},
                        "compose_type": "production",
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["source"], "pungi_cfg2#hash")

    def test_submit_build_unknown_source_type(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps({"source": {"type": "unknown", "source": "/path"}}),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["message"], 'Unknown source type "unknown"')

    def test_submit_module_build_wrong_source(self):
        with self.test_request_context(user="dev2"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "module", "source": "testmodule:master x"}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(
            data["message"],
            'Module definition must be in "n:s", "n:s:v" or '
            '"n:s:v:c" format, but got x',
        )

    def test_submit_module_build_base_module_in_source(self):
        with self.test_request_context(user="dev2"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {
                            "type": "module",
                            "source": "testmodule:master platform:x",
                        }
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(
            data["message"],
            "ODCS currently cannot create compose with base "
            "modules, but platform was requested.",
        )

    def test_submit_build_per_user_source_type_allowed(self):
        with self.test_request_context(user="dev2"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "module", "source": "testmodule:master"}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["state_name"], "wait")

    def test_submit_build_per_user_source_type_not_allowed(self):
        with self.test_request_context(user="dev2"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "tag", "source": "/path", "packages": ["foo"]}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(
            data["message"],
            "User dev2 not allowed to operate with compose with source_types=tag.",
        )

    def test_submit_build_per_group_source_type_allowed(self):
        with self.test_request_context(user="unknown", groups=["dev2", "x"]):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "module", "source": "testmodule:master"}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["state_name"], "wait")

    def test_submit_build_per_group_source_type_not_allowed(self):
        with self.test_request_context(user="unknown", groups=["dev2", "x"]):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "tag", "source": "/path", "packages": ["foo"]}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(
            data["message"],
            "User unknown not allowed to operate with compose with source_types=tag.",
        )

    def test_submit_scratch_build_tasks(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {
                            "type": "module",
                            "source": "testmodule:master",
                            "scratch_build_tasks": [
                                "123456",
                                "123457",
                            ],
                        }
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))
            self.assertEqual(
                data["scratch_build_tasks"],
                "123456 123457",
            )

    def test_submit_build_scratch_modules(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {
                            "type": "module",
                            "source": "testmodule:master",
                            "scratch_modules": [
                                "foo:bar:20200806:abcdefgh",
                                "fooo:bar:20200810:abcdefgh",
                            ],
                        }
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))
            self.assertEqual(
                data["source"],
                "testmodule:master",
            )
            self.assertEqual(
                data["scratch_modules"],
                "foo:bar:20200806:abcdefgh fooo:bar:20200810:abcdefgh",
            )

    def test_submit_build_only_scratch_modules(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {
                            "type": "module",
                            "source": "",
                            "scratch_modules": [
                                "foo:bar:20200806:abcdefgh",
                                "fooo:bar:20200810:abcdefgh",
                            ],
                        }
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))
            self.assertEqual(data["source"], "")
            self.assertEqual(
                data["scratch_modules"],
                "foo:bar:20200806:abcdefgh fooo:bar:20200810:abcdefgh",
            )

    def test_submit_build_parent_pungi_compose_ids(self):
        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "parent_pungi_compose_ids": ["Fedora-1-1", "Fedora-2-2"],
                        "source": {"type": "module", "source": "testmodule:master"},
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))
            self.assertEqual(
                data["parent_pungi_compose_ids"],
                "Fedora-1-1 Fedora-2-2",
            )

    def test_query_compose(self):
        resp = self.client.get("/api/1/composes/1")
        data = json.loads(resp.get_data(as_text=True))
        self.assertEqual(data["id"], 1)
        self.assertEqual(data["source"], "testmodule:master")
        self.assertEqual(data["pungi_config_dump"], None)
        self.assertEqual(resp.headers["Access-Control-Allow-Origin"], "*")

    def test_query_composes(self):
        resp = self.client.get("/api/1/composes/")
        evs = json.loads(resp.get_data(as_text=True))["items"]
        self.assertEqual(len(evs), 2)

    def test_query_compose_owner(self):
        resp = self.client.get("/api/1/composes/?owner=me")
        evs = json.loads(resp.get_data(as_text=True))["items"]
        self.assertEqual(len(evs), 1)
        self.assertEqual(evs[0]["source"], "f26")
        self.assertTrue("pungi_config_dump" not in evs[0])

    def test_query_compose_state_done(self):
        resp = self.client.get("/api/1/composes/?state=%d" % COMPOSE_STATES["done"])
        evs = json.loads(resp.get_data(as_text=True))["items"]
        self.assertEqual(len(evs), 0)

    def test_query_compose_state_wait(self):
        resp = self.client.get("/api/1/composes/?state=%d" % COMPOSE_STATES["wait"])
        evs = json.loads(resp.get_data(as_text=True))["items"]
        self.assertEqual(len(evs), 2)

    def test_query_compose_source_type(self):
        resp = self.client.get(
            "/api/1/composes/?source_type=%d" % PungiSourceType.MODULE
        )
        evs = json.loads(resp.get_data(as_text=True))["items"]
        self.assertEqual(len(evs), 1)

    def test_query_compose_source(self):
        resp = self.client.get("/api/1/composes/?source=f26")
        evs = json.loads(resp.get_data(as_text=True))["items"]
        self.assertEqual(len(evs), 1)

    def test_query_composes_order_by_default(self):
        resp = self.client.get("/api/1/composes/")
        composes = json.loads(resp.get_data(as_text=True))["items"]
        self.assertEqual([2, 1], [compose["id"] for compose in composes])

    def test_query_composes_order_by_id_asc(self):
        resp = self.client.get("/api/1/composes/?order_by=id")
        composes = json.loads(resp.get_data(as_text=True))["items"]
        self.assertEqual([1, 2], [compose["id"] for compose in composes])

    def test_query_composes_order_by_id_desc(self):
        resp = self.client.get("/api/1/composes/?order_by=-id")
        composes = json.loads(resp.get_data(as_text=True))["items"]
        self.assertEqual([2, 1], [compose["id"] for compose in composes])

    def test_query_composes_order_by_id_unknown_key(self):
        resp = self.client.get("/api/1/composes/?order_by=foo")
        data = json.loads(resp.get_data(as_text=True))
        self.assertEqual(data["status"], 400)
        self.assertEqual(data["error"], "Bad Request")
        self.assertTrue(
            data["message"].startswith(
                "An invalid order_by key was suplied, allowed keys are"
            )
        )

    def test_delete_compose(self):
        with freeze_time(self.initial_datetime) as frozen_datetime:
            c3 = Compose.create(
                db.session,
                "unknown",
                PungiSourceType.MODULE,
                "testmodule:master",
                COMPOSE_RESULTS["repository"],
                60,
            )
            c3.state = COMPOSE_STATES["done"]
            db.session.add(c3)
            db.session.commit()

            self.assertEqual(len(Compose.composes_to_expire()), 0)

            with self.test_request_context(user="root"):
                flask.g.oidc_scopes = [
                    "{0}{1}".format(conf.oidc_base_namespace, "delete-compose")
                ]

                resp = self.client.delete("/api/1/composes/%s" % c3.id)
                data = json.loads(resp.get_data(as_text=True))

            self.assertEqual(resp.status, "202 ACCEPTED")

            self.assertEqual(data["status"], 202)
            self.assertEqual(
                data["message"],
                "The delete request for compose (id=%s) has been accepted and will be processed by backend later."
                % c3.id,
            )

            self.assertEqual(c3.time_to_expire, self.initial_datetime)

            frozen_datetime.tick()
            self.assertEqual(len(Compose.composes_to_expire()), 1)
            expired_compose = Compose.composes_to_expire().pop()
            self.assertEqual(expired_compose.id, c3.id)
            self.assertEqual(expired_compose.removed_by, "root")

    def test_delete_not_allowed_states_compose(self):
        for state in COMPOSE_STATES.keys():
            if state not in ["wait", "done", "failed"]:
                new_c = Compose.create(
                    db.session,
                    "unknown",
                    PungiSourceType.MODULE,
                    "testmodule:master",
                    COMPOSE_RESULTS["repository"],
                    60,
                )
                new_c.state = COMPOSE_STATES[state]
                db.session.add(new_c)
                db.session.commit()
                compose_id = new_c.id

                with self.test_request_context(user="root"):
                    flask.g.oidc_scopes = [
                        "{0}{1}".format(conf.oidc_base_namespace, "delete-compose")
                    ]

                    resp = self.client.delete("/api/1/composes/%s" % compose_id)
                    data = json.loads(resp.get_data(as_text=True))

                self.assertEqual(resp.status, "400 BAD REQUEST")
                self.assertEqual(data["status"], 400)
                six.assertRegex(
                    self,
                    data["message"],
                    r"Compose \(id=%s\) can not be removed, its state need to be in .*."
                    % new_c.id,
                )
                self.assertEqual(data["error"], "Bad Request")

    def test_delete_non_exist_compose(self):
        with self.test_request_context(user="root"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "delete-compose")
            ]

            resp = self.client.delete("/api/1/composes/999999")
            data = json.loads(resp.get_data(as_text=True))

        self.assertEqual(resp.status, "404 NOT FOUND")
        self.assertEqual(data["status"], 404)
        self.assertEqual(data["message"], "No such compose found.")
        self.assertEqual(data["error"], "Not Found")

    def test_delete_compose_with_non_admin_user(self):
        self.c1.state = COMPOSE_STATES["failed"]
        db.session.commit()

        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "delete-compose")
            ]

            resp = self.client.delete("/api/1/composes/%s" % self.c1.id)
            data = json.loads(resp.get_data(as_text=True))

        self.assertEqual(resp.status, "403 FORBIDDEN")
        self.assertEqual(resp.status_code, 403)
        self.assertEqual(data["status"], 403)
        self.assertEqual(data["message"], "User dev is not in role admins.")

    def test_can_not_create_compose_with_non_composer_user(self):
        with self.test_request_context(user="qa"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            resp = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "module", "source": "testmodule:master"}}
                ),
            )
            data = json.loads(resp.get_data(as_text=True))

        self.assertEqual(resp.status, "403 FORBIDDEN")
        self.assertEqual(resp.status_code, 403)
        self.assertEqual(data["status"], 403)
        self.assertEqual(data["message"], "User qa is not in role allowed_clients.")

    def test_can_create_compose_with_user_in_configured_groups(self):
        with self.test_request_context(user="another_user", groups=["composer"]):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            resp = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "module", "source": "testmodule:rawhide"}}
                ),
            )
        db.session.expire_all()

        self.assertEqual(resp.status, "200 OK")
        self.assertEqual(resp.status_code, 200)
        c = (
            db.session.query(Compose)
            .filter(Compose.source == "testmodule:rawhide")
            .one()
        )
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

    def test_can_create_compose_with_user_in_multiple_groups(self):
        with self.test_request_context(user="another_user", groups=["dev3", "dev2"]):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            resp = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "module", "source": "testmodule:rawhide"}}
                ),
            )
        db.session.expire_all()

        self.assertEqual(resp.status, "200 OK")
        self.assertEqual(resp.status_code, 200)
        c = (
            db.session.query(Compose)
            .filter(Compose.source == "testmodule:rawhide")
            .one()
        )
        self.assertEqual(c.state, COMPOSE_STATES["wait"])

    def test_cannot_create_compose_with_user_in_multiple_groups(self):
        with self.test_request_context(user="another_user", groups=["dev3", "dev2"]):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            resp = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "tag", "source": "testmodule:rawhide"}}
                ),
            )
            data = json.loads(resp.get_data(as_text=True))
        db.session.expire_all()

        self.assertEqual(resp.status, "403 FORBIDDEN")
        self.assertEqual(resp.status_code, 403)
        self.assertEqual(
            data["message"],
            "User another_user not allowed to operate with compose with source_types=tag.",
        )

    def test_can_delete_compose_with_user_in_configured_groups(self):
        c3 = Compose.create(
            db.session,
            "unknown",
            PungiSourceType.MODULE,
            "testmodule:testbranch",
            COMPOSE_RESULTS["repository"],
            60,
        )
        c3.state = COMPOSE_STATES["done"]
        db.session.add(c3)
        db.session.commit()

        with self.test_request_context(user="another_admin", groups=["admin"]):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "delete-compose")
            ]

            resp = self.client.delete("/api/1/composes/%s" % c3.id)
            data = json.loads(resp.get_data(as_text=True))

        self.assertEqual(resp.status, "202 ACCEPTED")
        self.assertEqual(resp.status_code, 202)
        self.assertEqual(data["status"], 202)
        six.assertRegex(
            self,
            data["message"],
            r"The delete request for compose \(id=%s\) has been accepted and will be processed by backend later."
            % c3.id,
        )

    def test_can_create_compose_with_permission_overriden_by_username(self):
        with self.test_request_context(user="dev3", groups=["dev2"]):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            resp = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "module", "source": "testmodule:rawhide"}}
                ),
            )
        db.session.expire_all()

        self.assertEqual(resp.status, "403 FORBIDDEN")
        self.assertEqual(resp.status_code, 403)

    @patch.object(
        odcs.server.config.Config, "max_seconds_to_live", new_callable=PropertyMock
    )
    @patch.object(
        odcs.server.config.Config, "seconds_to_live", new_callable=PropertyMock
    )
    def test_use_seconds_to_live_in_request(
        self, mock_seconds_to_live, mock_max_seconds_to_live
    ):
        # Test that seconds-to-live is still supported to keep backward compatibility.
        for seconds_to_live in ["seconds-to-live", "seconds_to_live"]:
            # if we have 'seconds_to_live' in request < conf.max_seconds_to_live
            # the value from request will be used
            mock_seconds_to_live.return_value = 60 * 60 * 24
            mock_max_seconds_to_live.return_value = 60 * 60 * 24 * 3

            with self.test_request_context(user="dev"):
                flask.g.oidc_scopes = [
                    "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
                ]

                rv = self.client.post(
                    "/api/1/composes/",
                    data=json.dumps(
                        {
                            "source": {"type": "module", "source": "testmodule:master"},
                            "seconds_to_live": 60 * 60 * 12,
                        }
                    ),
                )
                data = json.loads(rv.get_data(as_text=True))

            time_submitted = datetime.strptime(
                data["time_submitted"], "%Y-%m-%dT%H:%M:%SZ"
            )
            time_to_expire = datetime.strptime(
                data["time_to_expire"], "%Y-%m-%dT%H:%M:%SZ"
            )
            delta = timedelta(hours=12)
            self.assertEqual(time_to_expire - time_submitted, delta)

    @patch.object(
        odcs.server.config.Config, "max_seconds_to_live", new_callable=PropertyMock
    )
    @patch.object(
        odcs.server.config.Config, "seconds_to_live", new_callable=PropertyMock
    )
    def test_use_max_seconds_to_live_in_conf(
        self, mock_seconds_to_live, mock_max_seconds_to_live
    ):
        # if we have 'seconds_to_live' in request > conf.max_seconds_to_live
        # conf.max_seconds_to_live will be used
        mock_seconds_to_live.return_value = 60 * 60 * 24
        mock_max_seconds_to_live.return_value = 60 * 60 * 24 * 3

        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "module", "source": "testmodule:master"},
                        "seconds_to_live": 60 * 60 * 24 * 7,
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        time_submitted = datetime.strptime(data["time_submitted"], "%Y-%m-%dT%H:%M:%SZ")
        time_to_expire = datetime.strptime(data["time_to_expire"], "%Y-%m-%dT%H:%M:%SZ")
        delta = timedelta(days=3)
        self.assertEqual(time_to_expire - time_submitted, delta)

    @patch.object(
        odcs.server.config.Config, "max_seconds_to_live", new_callable=PropertyMock
    )
    @patch.object(
        odcs.server.config.Config, "seconds_to_live", new_callable=PropertyMock
    )
    def test_use_seconds_to_live_in_conf(
        self, mock_seconds_to_live, mock_max_seconds_to_live
    ):
        # if we don't have 'seconds_to_live' in request, conf.seconds_to_live will be used
        mock_seconds_to_live.return_value = 60 * 60 * 24
        mock_max_seconds_to_live.return_value = 60 * 60 * 24 * 3

        with self.test_request_context(user="dev"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "module", "source": "testmodule:master"}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        time_submitted = datetime.strptime(data["time_submitted"], "%Y-%m-%dT%H:%M:%SZ")
        time_to_expire = datetime.strptime(data["time_to_expire"], "%Y-%m-%dT%H:%M:%SZ")
        delta = timedelta(hours=24)
        self.assertEqual(time_to_expire - time_submitted, delta)

    @patch.object(odcs.server.config.Config, "auth_backend", new_callable=PropertyMock)
    def test_anonymous_user_can_submit_build_with_noauth_backend(
        self, mock_auth_backend
    ):
        mock_auth_backend.return_value = "noauth"

        with self.test_request_context():
            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "module", "source": "testmodule:master"}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        expected_json = {
            "source_type": 2,
            "state": 0,
            "time_done": None,
            "state_name": "wait",
            "state_reason": None,
            "source": u"testmodule:master",
            "owner": u"unknown",
            "result_repo": "http://localhost/odcs/odcs-%d/compose/Temporary"
            % data["id"],
            "result_repofile": "http://localhost/odcs/odcs-%d/compose/Temporary/odcs-%d.repo"
            % (data["id"], data["id"]),
            "time_submitted": data["time_submitted"],
            "id": data["id"],
            "time_started": None,
            "time_removed": None,
            "removed_by": None,
            "time_to_expire": data["time_to_expire"],
            "flags": [],
            "results": ["repository"],
            "sigkeys": "",
            "koji_event": None,
            "koji_task_id": None,
            "packages": None,
            "builds": None,
            "arches": "x86_64",
            "multilib_arches": "",
            "multilib_method": 0,
            "lookaside_repos": "",
            "modular_koji_tags": None,
            "module_defaults_url": None,
            "label": None,
            "compose_type": "test",
            "pungi_compose_id": None,
            "target_dir": "default",
            "scratch_modules": None,
            "toplevel_url": "http://localhost/odcs/odcs-%d" % data["id"],
            "parent_pungi_compose_ids": None,
            "scratch_build_tasks": None,
        }
        self.assertEqual(data, expected_json)

        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])


class TestExtendExpiration(ViewBaseTest):
    """Test view post to extend expiration"""

    def setUp(self):
        super(TestExtendExpiration, self).setUp()
        self.oidc_base_namespace = patch.object(
            conf, "oidc_base_namespace", new="http://example.com/"
        )
        self.oidc_base_namespace.start()

    def tearDown(self):
        self.oidc_base_namespace.stop()
        super(TestExtendExpiration, self).tearDown()

    def setup_test_data(self):
        self.initial_datetime = datetime(
            year=2016, month=1, day=1, hour=0, minute=0, second=0
        )
        with freeze_time(self.initial_datetime):
            self.c1 = Compose.create(
                db.session,
                "me",
                PungiSourceType.KOJI_TAG,
                "f25",
                COMPOSE_RESULTS["repository"],
                60,
            )
            self.c2 = Compose.create(
                db.session,
                "me",
                PungiSourceType.KOJI_TAG,
                "f26",
                COMPOSE_RESULTS["repository"],
                60,
            )
            self.c3 = Compose.create(
                db.session,
                "me",
                PungiSourceType.KOJI_TAG,
                "f27",
                COMPOSE_RESULTS["repository"],
                60,
            )
            self.c4 = Compose.create(
                db.session,
                "me",
                PungiSourceType.KOJI_TAG,
                "master",
                COMPOSE_RESULTS["repository"],
                60,
            )

            map(db.session.add, (self.c1, self.c2, self.c3, self.c4))
            db.session.commit()

            self.c1.reused_id = self.c2.id
            self.c2.reused_id = self.c3.id
            db.session.commit()

            # Store for retrieving them back for assertion
            self.c1_id = self.c1.id
            self.c3_id = self.c3.id

    @patch.object(conf, "auth_backend", new="noauth")
    def test_bad_request_if_seconds_to_live_is_invalid(self):
        post_data = json.dumps({"seconds_to_live": "600s"})
        with self.test_request_context():
            rv = self.client.patch(
                "/api/1/composes/{0}".format(self.c1.id), data=post_data
            )
            data = json.loads(rv.get_data(as_text=True))

            self.assertEqual(400, data["status"])
            self.assertEqual("Bad Request", data["error"])
            self.assertIn(
                "Invalid seconds_to_live specified in request", data["message"]
            )

    @patch.object(conf, "auth_backend", new="noauth")
    def test_bad_request_if_request_data_is_not_json(self):
        with self.test_request_context():
            rv = self.client.patch("/api/1/composes/{0}".format(self.c1.id), data="abc")
            data = json.loads(rv.get_data(as_text=True))

            self.assertEqual(400, data["status"])
            self.assertEqual("Bad Request", data["error"])
            self.assertIn("Failed to decode JSON object", data["message"])

    @patch.object(conf, "oidc_base_namespace", new="http://example.com/")
    def test_fail_if_extend_non_existing_compose(self):
        post_data = json.dumps({"seconds_to_live": 600})
        with self.test_request_context():
            flask.g.oidc_scopes = [
                "http://example.com/new-compose",
                "http://example.com/renew-compose",
            ]

            rv = self.client.patch("/api/1/composes/999", data=post_data)
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual("No compose with id 999 found", data["message"])

    def test_fail_if_compose_is_not_done(self):
        self.c1.state = COMPOSE_STATES["wait"]
        db.session.commit()

        post_data = json.dumps({"seconds_to_live": 600})
        with self.test_request_context():
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "renew-compose")
            ]

            rv = self.client.patch(
                "/api/1/composes/{0}".format(self.c1.id), data=post_data
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(
            "No compose with id {0} found".format(self.c1.id), data["message"]
        )

    def test_extend_compose_expiration(self):
        fake_utcnow = datetime.utcnow()

        self.c2.state = COMPOSE_STATES["done"]
        self.c2.time_to_expire = fake_utcnow - timedelta(seconds=10)
        db.session.commit()

        expected_seconds_to_live = 60 * 60 * 3
        expected_time_to_expire = fake_utcnow + timedelta(
            seconds=expected_seconds_to_live
        )
        post_data = json.dumps({"seconds_to_live": expected_seconds_to_live})

        with self.test_request_context():
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "renew-compose")
            ]
            with freeze_time(fake_utcnow):
                url = "/api/1/composes/{0}".format(self.c2.id)
                rv = self.client.patch(url, data=post_data)
                data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(
            Compose._utc_datetime_to_iso(expected_time_to_expire),
            data["time_to_expire"],
        )

        # Compose reusing self.c2 and the one self.c2 reuses should also be
        # extended.
        c1 = db.session.query(Compose).filter(Compose.id == self.c1_id).first()
        self.assertEqual(expected_time_to_expire, c1.time_to_expire)
        c3 = db.session.query(Compose).filter(Compose.id == self.c3_id).first()
        self.assertEqual(expected_time_to_expire, c3.time_to_expire)


class TestViewsRawConfig(ViewBaseTest):
    maxDiff = None

    def setUp(self):
        super(TestViewsRawConfig, self).setUp()
        self.oidc_base_namespace = patch.object(
            conf, "oidc_base_namespace", new="http://example.com/"
        )
        self.oidc_base_namespace.start()

    def tearDown(self):
        super(TestViewsRawConfig, self).tearDown()
        self.oidc_base_namespace.stop()

    def test_submit_build_raw_config_too_many_sources(self):
        with self.test_request_context(user="dev2"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {
                            "type": "raw_config",
                            "source": "pungi_cfg#hash pungi2cfg_hash",
                        }
                    }
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["status"], 400)
        self.assertEqual(
            data["message"],
            'Only single source is allowed for "raw_config" source_type',
        )

    def test_submit_build_raw_config_no_hash(self):
        with self.test_request_context(user="dev2"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "raw_config", "source": "pungi_cfg"}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["status"], 400)
        self.assertEqual(
            data["message"],
            'Source must be in "source_name#commit_hash" format for '
            '"raw_config" source_type.',
        )

    def test_submit_build_raw_config_empty_hash(self):
        with self.test_request_context(user="dev2"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "raw_config", "source": "pungi_cfg#"}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["status"], 400)
        self.assertEqual(
            data["message"],
            'Source must be in "source_name#commit_hash" format for '
            '"raw_config" source_type.',
        )

    def test_submit_build_raw_config_empty_name(self):
        with self.test_request_context(user="dev2"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps({"source": {"type": "raw_config", "source": "#hash"}}),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["status"], 400)
        self.assertEqual(
            data["message"],
            'Source must be in "source_name#commit_hash" format for '
            '"raw_config" source_type.',
        )

    def test_submit_build_raw_config_empty_name_hash(self):
        with self.test_request_context(user="dev2"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps({"source": {"type": "raw_config", "source": "#"}}),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["status"], 400)
        self.assertEqual(
            data["message"],
            'Source must be in "source_name#commit_hash" format for '
            '"raw_config" source_type.',
        )

    def test_submit_build_raw_config_unknown_name(self):
        with self.test_request_context(user="dev2"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            rv = self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {"source": {"type": "raw_config", "source": "pungi_cfg#hash"}}
                ),
            )
            data = json.loads(rv.get_data(as_text=True))

        self.assertEqual(data["status"], 400)
        self.assertEqual(
            data["message"],
            'Source "pungi_cfg" does not exist in server configuration.',
        )

    @patch.object(
        odcs.server.config.Config,
        "raw_config_urls",
        new={"pungi_cfg": "http://localhost/pungi.conf#%s"},
    )
    def test_submit_build_raw_config(self):
        with self.test_request_context(user="dev2"):
            flask.g.oidc_scopes = [
                "{0}{1}".format(conf.oidc_base_namespace, "new-compose")
            ]

            self.client.post(
                "/api/1/composes/",
                data=json.dumps(
                    {
                        "source": {"type": "raw_config", "source": "pungi_cfg#hash"},
                        "label": "Beta-1.2",
                        "compose_type": "nightly",
                    }
                ),
            )
        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.state, COMPOSE_STATES["wait"])
        self.assertEqual(c.source_type, PungiSourceType.RAW_CONFIG)
        self.assertEqual(c.source, "pungi_cfg#hash")
        self.assertEqual(c.label, "Beta-1.2")
        self.assertEqual(c.compose_type, "nightly")
        self.assertEqual(c.json()["result_repofile"], "")


class TestViewsCancelCompose(ViewBaseTest):
    maxDiff = None

    def setup_test_data(self):
        self.initial_datetime = datetime(
            year=2016, month=1, day=1, hour=0, minute=0, second=0
        )
        with freeze_time(self.initial_datetime):
            self.c1 = Compose.create(
                db.session,
                "dev2",
                PungiSourceType.MODULE,
                "testmodule:master",
                COMPOSE_RESULTS["repository"],
                60,
            )
            db.session.commit()
            self.task_id = "71267f28-5194-4720-b57b-a665fabdb012"
            self.c1.celery_task_id = self.task_id
            db.session.commit()
            self.c1_id = self.c1.id

    @patch("odcs.server.views.CELERY_AVAILABLE", new=False)
    def test_no_celery(self):
        with self.test_request_context(user="dev2"):
            resp = self.client.delete("/api/1/composes/%s" % self.c1.id)
        # Without celery we can't cancel, so the code should try to delete the
        # compose and fail on user not being an admin.
        self.assertEqual(resp.status_code, 403)
        data = json.loads(resp.get_data(as_text=True))
        self.assertEqual(data["status"], 403)
        self.assertEqual(data["message"], "User dev2 is not in role admins.")

    def test_bad_owner(self):
        with self.test_request_context(user="dev1"):
            resp = self.client.delete("/api/1/composes/%s" % self.c1.id)
        self.assertEqual(resp.status_code, 403)
        data = json.loads(resp.get_data(as_text=True))
        self.assertEqual(data["status"], 403)
        self.assertEqual(
            data["message"],
            "Compose (id=%s) can not be canceled, it's owned by someone else.",
        )

    @patch("odcs.server.views.celery_app")
    def test_cancel(self, app):
        with self.test_request_context(user="dev2"):
            resp = self.client.delete("/api/1/composes/%s" % self.c1_id)
        self.assertEqual(resp.status_code, 202)
        self.assertEqual(app.mock_calls, [call.control.revoke(self.task_id)])
        data = json.loads(resp.get_data(as_text=True))
        self.assertEqual(data["status"], 202)
        self.assertEqual(
            data["message"], "Compose (id=%s) has been canceled" % self.c1_id
        )
        c1 = db.session.query(Compose).filter(Compose.id == self.c1_id).first()
        self.assertEqual(c1.state, COMPOSE_STATES["failed"])
        self.assertEqual(c1.state_reason, "Canceled by dev2")



# -*- coding: utf-8 -*-
#
# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time
import unittest
from mock import patch, mock_open, MagicMock

from odcs.server.mock_runroot import (
    mock_runroot_init,
    raise_if_runroot_key_invalid,
    mock_runroot_run,
    mock_runroot_install,
    rmtree_skip_mounts,
    cleanup_old_runroots,
)
from .utils import AnyStringWith


class TestMockRunroot(unittest.TestCase):
    def setUp(self):
        super(TestMockRunroot, self).setUp()

    def tearDown(self):
        super(TestMockRunroot, self).tearDown()

    @patch("odcs.server.mock_runroot.create_koji_session")
    @patch("odcs.server.mock_runroot.execute_mock")
    @patch("odcs.server.mock_runroot.print", create=True)
    @patch("odcs.server.mock_runroot.rmtree_skip_mounts")
    @patch("odcs.server.mock_runroot.cleanup_old_runroots")
    def test_mock_runroot_init(
        self,
        cleanup_old_runroots,
        rmtree_skip_mounts,
        fake_print,
        execute_mock,
        create_koji_session,
    ):
        execute_mock.side_effect = RuntimeError("Expected exception")
        koji_session = create_koji_session.return_value
        koji_session.getRepo.return_value = {"id": 1}

        m = mock_open()
        with patch("odcs.server.mock_runroot.open", m, create=True):
            with self.assertRaises(RuntimeError):
                mock_runroot_init("f30-build")

        fake_print.assert_called_once()
        m.return_value.write.assert_called_once_with(AnyStringWith("f30-build"))

        execute_mock.assert_called_once_with(AnyStringWith("-"), ["--init"])
        rmtree_skip_mounts.assert_called_once()
        cleanup_old_runroots.assert_called_once()

    def test_raise_if_runroot_key_invalid(self):
        with self.assertRaises(ValueError):
            raise_if_runroot_key_invalid("../../test")
        with self.assertRaises(ValueError):
            raise_if_runroot_key_invalid("/tmp")
        with self.assertRaises(ValueError):
            raise_if_runroot_key_invalid("x.cfg")
        raise_if_runroot_key_invalid("1-2-3-4-a-s-d-f")

    @patch("odcs.server.mock_runroot.execute_mock")
    @patch("odcs.server.mock_runroot.execute_cmd")
    def test_mock_runroot_run(self, execute_cmd, execute_mock):
        mock_runroot_run("foo-bar", ["df", "-h"])

        execute_mock.assert_called_once_with(
            "foo-bar",
            ["--old-chroot", "--chroot", "--", "/bin/sh", "-c", "{ df -h; }"],
            False,
        )
        execute_cmd.assert_any_call(
            [
                "mount",
                "-o",
                "bind",
                AnyStringWith("test_composes"),
                AnyStringWith("test_composes"),
            ]
        )
        execute_cmd.assert_any_call(["umount", "-l", AnyStringWith("test_composes")])

    @patch("odcs.server.mock_runroot.execute_mock")
    def test_mock_runroot_install(self, execute_mock):
        mock_runroot_install("foo-bar", ["lorax", "dracut"])
        execute_mock.assert_called_once_with(
            "foo-bar", ["--install", "lorax", "dracut"]
        )

    @patch("odcs.server.mock_runroot.execute_mock")
    @patch("odcs.server.mock_runroot.rmtree_skip_mounts")
    def test_mock_runroot_install_exception(self, rmtree_skip_mounts, execute_mock):
        execute_mock.side_effect = RuntimeError("Expected exception")
        with self.assertRaises(RuntimeError):
            mock_runroot_install("foo-bar", ["lorax", "dracut"])
        rmtree_skip_mounts.assert_called_once()

    @patch("odcs.server.mock_runroot.execute_mock")
    @patch("odcs.server.mock_runroot.execute_cmd")
    @patch("odcs.server.mock_runroot.rmtree_skip_mounts")
    def test_mock_runroot_run_exception(
        self, rmtree_skip_mounts, execute_cmd, execute_mock
    ):
        execute_mock.side_effect = RuntimeError("Expected exception")
        with self.assertRaises(RuntimeError):
            mock_runroot_run("foo-bar", ["df", "-h"])

        execute_mock.assert_called_once_with(
            "foo-bar",
            ["--old-chroot", "--chroot", "--", "/bin/sh", "-c", "{ df -h; }"],
            False,
        )
        execute_cmd.assert_any_call(
            [
                "mount",
                "-o",
                "bind",
                AnyStringWith("test_composes"),
                AnyStringWith("test_composes"),
            ]
        )
        execute_cmd.assert_any_call(["umount", "-l", AnyStringWith("test_composes")])
        rmtree_skip_mounts.assert_called_once()

    @patch("odcs.server.mock_runroot.os.rmdir")
    @patch("odcs.server.mock_runroot.os.listdir")
    @patch("odcs.server.mock_runroot.os.lstat")
    @patch("odcs.server.mock_runroot.os.remove")
    @patch("odcs.server.mock_runroot.stat.S_ISDIR")
    def test_mock_runroot_rmtree(self, isdir, remove, lstat, listdir, rmdir):
        """
        Tests that `rmtree_skip_mounts` really skips the mount points when
        removing the runroot root directory.

        The fake runroot root directory in this test is following:
        - /mnt/koji - empty directory which should be removed
        - /mnt/odcs - non-empty mountpoint directory must not be removed.
        - /x - regular file which should be removed.
        """

        def mocked_listdir(path):
            # Creates fake directory structure within the /var/lib/mock/foo-bar/root:
            #  - /mnt/koji
            #  - /mnt/odcs/foo
            #  - /x
            if path == "/var/lib/mock/foo-bar/root":
                return ["mnt", "x"]
            if path.endswith("/mnt"):
                return ["odcs", "koji"]
            elif path.endswith("/odcs"):
                return ["foo"]
            return []

        listdir.side_effect = mocked_listdir

        def mocked_isdir(mode):
            # We use fake values here:
            # - 0 means it is not a directory.
            # - 1 means it is a directory.
            return mode

        isdir.side_effect = mocked_isdir

        def mocked_lstat(path):
            stat_result = MagicMock()
            if path.endswith("/x"):
                stat_result.st_mode = 0  # Just fake return value for regular file.
            else:
                stat_result.st_mode = 1  # Fake value for directory.
            return stat_result

        lstat.side_effect = mocked_lstat

        rmtree_skip_mounts("/var/lib/mock/foo-bar/root", ["/mnt/odcs"])

        rmdir.assert_called_once_with("/var/lib/mock/foo-bar/root/mnt/koji")
        remove.assert_called_once_with("/var/lib/mock/foo-bar/root/x")

    @patch("odcs.server.mock_runroot.os.listdir")
    @patch("odcs.server.mock_runroot.os.stat")
    @patch("odcs.server.mock_runroot.rmtree_skip_mounts")
    def test_cleanup_old_runroot(self, rmtree_skip_mounts, stat, listdir):
        listdir.return_value = ["foo", "bar", "already-removed"]

        def mocked_stat(path):
            stat_result = MagicMock()
            if path.endswith("/foo"):
                # The "foo/root" is 1 day old, so should be removed.
                stat_result.st_mtime = time.time() - 24 * 3600
            elif path.endswith("/already-removed"):
                # The "already-removed/root" is already removed, so raise an
                # exception.
                raise OSError("No such file")
            elif path.endswith("/bar"):
                # The "bar/root" is just 10 seconds old, so should not be
                # removed.
                stat_result.st_mtime = time.time() - 10
            else:
                raise ValueError("stat called for unexpected file.")
            return stat_result

        stat.side_effect = mocked_stat

        cleanup_old_runroots()

        rmtree_skip_mounts.assert_called_once_with(
            "/var/lib/mock/foo", AnyStringWith("test_composes")
        )



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import os
import requests
from xml.etree import ElementTree
from concurrent.futures import ThreadPoolExecutor
from datetime import timedelta
from six.moves.urllib.parse import urlparse
from distutils.spawn import find_executable

from flufl.lock import Lock

from odcs.server import log, conf
from odcs.server.utils import makedirs, execute_cmd, retry


class MergeRepo(object):
    def __init__(self, compose):
        self.compose = compose

    @retry(wait_on=(requests.ConnectionError,), logger=log)
    def _download_file(self, path, url):
        """
        Downloads repodata file, stores it into `path`/repodata and returns
        its content.

        :param str path: Path to store the file to.
        :param str url: URL of file to download.
        :rtype: str
        :return: path to the downloaded file
        """
        log.info("%r: Downloading %s", self.compose, url)
        r = requests.get(url, timeout=conf.net_timeout, stream=True)
        r.raise_for_status()

        filename = os.path.basename(url)
        makedirs(os.path.join(path, "repodata"))
        outfile = os.path.join(path, "repodata", filename)
        with open(outfile, "wb") as f:
            for chunk in r.iter_content(chunk_size=None):
                f.write(chunk)
        response_len = os.stat(outfile).st_size
        log.info("%r: Downloaded %d bytes from %s", self.compose, response_len, url)
        return outfile

    def _download_repodata(self, path, baseurl):
        """
        Downloads the repodata from `baseurl` to `path`.

        :param str path: Path to store the file to.
        :param str baseurl: Base URL of repository to download.
        """
        # In case we already have the repodata downloaded, read the repomd.xml,
        # so we can later check if the repodata changed or not.
        last_repomd = None
        repodata_path = os.path.join(path, "repodata")
        repomd_path = os.path.join(repodata_path, "repomd.xml")
        if os.path.exists(repomd_path):
            with open(repomd_path, "r") as f:
                last_repomd = f.read()

        # Download the repomd.xml.
        repomd_url = os.path.join(baseurl, "repodata", "repomd.xml")
        repomd_path = self._download_file(path, repomd_url)
        with open(repomd_path) as f:
            repomd = f.read()
        tree = ElementTree.fromstring(repomd)

        # In case the repomd.xml did not change since the last compose, use
        # the existing repodata to save lot of time downloading other repodata
        # files.
        if repomd == last_repomd:
            log.info("%r: Reusing cached repodata for %s", self.compose, baseurl)
            return

        # In case the repomd.xml changed, remove everything from the repodata
        # directory except the just downloaded repomd.xml.
        for f in os.listdir(repodata_path):
            if f == "repomd.xml":
                continue
            os.unlink(os.path.join(repodata_path, f))

        # Parse the repomd.xml and download all the repodata files needed
        # to merge the repos.
        ns = "{http://linux.duke.edu/metadata/repo}"
        with ThreadPoolExecutor(5) as downloader:
            for data in tree.findall("%sdata" % ns):
                if data.get("type").endswith("_db"):
                    continue
                data_location = data.find("%slocation" % ns).get("href")
                data_url = os.path.join(baseurl, data_location)
                downloader.submit(self._download_file, path, data_url)

    def run(self, arch, repos, repo_name):
        """
        Merges multiple RPM repositories and stores the output to
        `os.path.join(compose.result_repo_dir, repo_name, arch)`.

        Raises an RuntimeError in case of error.

        :param str arch: Architecture of RPMs in repositories.
        :param list repos: List of URLs pointing to repos to merge.
        :param str repo_name: Name of the repository defining the subdirectory
            in `compose.result_repo_dir`.
        """
        # Multiple MergeRepo tasks can be running in the same time, so the
        # pulp_repo_cache must be protected by lock.
        # This list holds the lock for each pulp repo we merge.
        locks = []

        # Contains paths to per pulp repo pulp_repo_cache sub-directories.
        repo_paths = []

        # Remove duplicated URLs from repos. It is useless to merge the same URLs and it would
        # also break the locking code which tries to lock the same cache directory twice.
        # The mergerepo_c can handle the case when we end up with just single repo in the `repos`.
        repos = list(set(repos))
        repos.sort()

        parsed_url = urlparse(repos[0])
        repo_prefix = "%s://%s" % (parsed_url.scheme, parsed_url.hostname)
        repo_prefix = repo_prefix.strip("/") + "/"

        # Generate the pulp_repo_cache structure and locks for each repo.
        for repo in repos:
            repo_path = os.path.join(
                self.compose.target_dir,
                "pulp_repo_cache",
                repo.replace(repo_prefix, ""),
            )
            repo_paths.append(repo_path)
            makedirs(repo_path)

            lock = Lock(os.path.join(repo_path, "lock"))
            lock.lifetime = timedelta(minutes=30)
            locks.append(lock)
        try:
            # Lock the locks and download the repodata.
            for lock in locks:
                lock.lock()
            for repo, repo_path in zip(repos, repo_paths):
                self._download_repodata(repo_path, repo)

            log.info("%r: Starting mergerepo_c: %r", self.compose, repo_paths)
            mergerepo_exe = find_executable("mergerepo_c")
            if not mergerepo_exe:
                raise RuntimeError("mergerepo_c is not available on system")

            result_repo_dir = os.path.join(
                self.compose.result_repo_dir, repo_name, arch
            )
            makedirs(result_repo_dir)

            args = [mergerepo_exe, "--method", "nvr", "-o", result_repo_dir]
            args += [
                "--repo-prefix-search",
                os.path.join(self.compose.target_dir, "pulp_repo_cache"),
            ]
            args += ["--repo-prefix-replace", repo_prefix]
            for repo in repo_paths:
                args.append("-r")
                args.append(repo)

            execute_cmd(args, timeout=conf.mergerepo_timeout)
        finally:
            for lock in locks:
                if lock.is_locked:
                    lock.unlock()



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>


from threading import Lock
from sqlalchemy.orm import attributes

from logging import getLogger

log = getLogger()

_cache_lock = Lock()
_cached_composes = {}


def cache_composes_if_state_changed(session, flush_context):
    """Prepare outgoing messages when compose state is changed"""

    from odcs.server.models import Compose

    composes = (
        item for item in (session.new | session.dirty) if isinstance(item, Compose)
    )
    composes_state_changed = (
        compose
        for compose in composes
        if not attributes.get_history(compose, "state").unchanged
    )

    with _cache_lock:
        for comp in composes_state_changed:
            if comp.id not in _cached_composes:
                _cached_composes[comp.id] = []
            _cached_composes[comp.id].append(comp.json())

    log.debug(
        "Cached composes to be sent due to state changed: %s", _cached_composes.keys()
    )


def start_to_publish_messages(session):
    """Publish messages after data is committed to database successfully"""
    import odcs.server.messaging as messaging

    with _cache_lock:
        msgs = []
        for compose_jsons in _cached_composes.values():
            for compose_json in compose_jsons:
                msgs.append({"event": "state-changed", "compose": compose_json})
        log.debug("Sending messages: %s", msgs)
        if msgs:
            try:
                messaging.publish(msgs)
            except Exception:
                log.exception("Cannot publish message to bus.")
        _cached_composes.clear()



# -*- coding: utf-8 -*-
# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>
#
# This script is executed on runroot hosts when Pungi needs to execute runroot
# tasks. It is therefore execute with "root" permissions, initializes the
# Mock environment and allows installing packages and running commands in
# the Mock environment.
#
# The script is executed using SSH from the ODCS backend.
#
# The workflow is following:
#
#   1) The `mock_runroot_init` is called. This generates unique ID defining
#      the Mock chroot. It initializes the Mock environment and prints
#      the unique ID ("runroot_key") to stdout. The "runroot_key" is later
#      used to identify the Mock chroot.
#   2) The `mock_runroot_install` is called to install the requested packages
#      in the Mock chroot for the runroot task. The "runroot_key" is used
#      to get the particular Mock chroot.
#   3) The `mock_runroot_run` is called to run the runroot command in
#      the Mock chroot. The conf.target_dir is mounted in the Mock chroot, so
#      the output of this command can be stored there.

from __future__ import print_function
import time
import platform
import sys
import koji
import os
import uuid
import tempfile
import logging
import stat

from odcs.server import conf
from odcs.server.backend import create_koji_session
from odcs.server.utils import makedirs, execute_cmd


def do_mounts(rootdir, mounts):
    """
    Mounts the host `mounts` in the Mock chroot `rootdir`.

    :param str rootdir: Full path to root directory for Mock chroot.
    :param list mounts: Full paths to mount directories which will be mounted
        in the `rootdir`.
    """
    for mount in mounts:
        mpoint = "%s%s" % (rootdir, mount)
        makedirs(mpoint)
        cmd = ["mount", "-o", "bind", mount, mpoint]
        execute_cmd(cmd)


def undo_mounts(rootdir, mounts):
    """
    Umounts the host `mounts` from the Mock chroot `rootdir`.

    :param str rootdir: Full path to root directory for Mock chroot.
    :param list mounts: Full paths to mount directories which will be umounted
        from the `rootdir`.
    """
    for mount in mounts:
        mpoint = "%s%s" % (rootdir, mount)
        cmd = ["umount", "-l", mpoint]
        execute_cmd(cmd)


def rmtree_skip_mounts(rootdir, mounts, rootdir_mounts=None):
    """
    The rmtree method based on `shutil.rmtree` skipping the `mounts` directories.
    We want to skip these directories in case the umount fails for whatever reason.
    We need to ensure that /mnt/odcs content is not removed.

    This method catches the `os.error` exceptions, so the runroot task does
    not fail because of rmtree error.

    :param str rootdir: Full path to root directory for Mock chroot to remove.
        For example "/var/lib/mock/foo/root".
    :param list mounts: List of mount points to skip in case they are not umounted.
        For example ["/mnt/odcs", "/mnt/koji"].
    :param str rootdir_mounts: Helper variable for recursive calls of this
        function. It is initialized by this function in the first call and is
        passed to recursive calls. It contains the full-path to mount point.
        For example:
            ["/var/lib/mock/foo/root/mnt/odcs", "/var/lib/mock/foo/root/mnt/koji"]
    :return bool: True if some subdirectory of `rootdir` has been skipped and
        not removed.
    """
    if not rootdir_mounts:
        rootdir_mounts = ["%s%s" % (rootdir, mount) for mount in mounts]

    # Skip the directory which is mount point in case it contains some files -
    # that means it has not been umounted.
    # This counts with recursive call of this method. For example:
    #   rootdir_mounts = ["/var/lib/mock/foo/root/mnt/koji"]
    #   Call #1:      rootdir = "/var/lib/mock/foo/root/"
    #     Call #2:    rootdir = "/var/lib/mock/foo/root/mnt"
    #       Call #3:  rootdir = "/var/lib/mock/foo/root/mnt/koji"
    # In the Call #3, the rootdir == rootdir_mounts[0], so this directory is
    # not removed when not empty and True is returned back to calls #2 and #1.
    # That tells the #2 and #1 to also not remove the "rootdir" with which
    # they have been called.
    if rootdir in rootdir_mounts and os.listdir(rootdir):
        return True

    subdirectory_skipped = False
    names = []
    try:
        names = os.listdir(rootdir)
    except os.error:
        pass
    for name in names:
        fullname = os.path.join(rootdir, name)
        try:
            mode = os.lstat(fullname).st_mode
        except os.error:
            mode = 0
        if stat.S_ISDIR(mode):
            if rmtree_skip_mounts(fullname, mounts, rootdir_mounts):
                subdirectory_skipped = True
        else:
            try:
                os.remove(fullname)
            except os.error:
                continue
    if not subdirectory_skipped:
        try:
            os.rmdir(rootdir)
        except os.error:
            pass
    return subdirectory_skipped


def cleanup_old_runroots():
    """
    Checks the /var/lib/mock directory for old runroot chroots and remove them.

    Those chroots can be lost there for example when the host reboots in the
    middle of runroot generation or in case Pungi task gets killed for
    whatever reason.

    We need to ensure that the chroot data are removed from the filesystem
    in these cases.
    """
    now = time.time()
    mounts = [conf.target_dir] + conf.runroot_extra_mounts
    mock_root = "/var/lib/mock"
    for runroot_key in os.listdir(mock_root):
        mock_dir = os.path.join(mock_root, runroot_key)
        # Skip the mock_dir if it is not old enough.
        try:
            if os.stat(mock_dir).st_mtime > now - conf.pungi_timeout:
                continue
        except OSError:
            continue
        rmtree_skip_mounts(mock_dir, mounts)


def runroot_tmp_path(runroot_key):
    """
    Creates and returns the temporary path to store the configuration files
    or logs for the runroot task.

    :param str runroot_key: The Runroot key.
    :return str: Full-path to temporary directory.
    """
    path = os.path.join(tempfile.gettempdir(), "odcs-runroot-%s" % runroot_key)
    makedirs(path)
    return path


def execute_mock(runroot_key, args, log_output=True):
    """
    Executes the Mock command with `args` for given `runroot_key` Mock chroot.

    :param str runroot_key: Runroot key.
    :param list args: The "mock" command arguments.
    :param bool log_output: When True, stdout and stderr of Mock command are
        redirected to logs.
    """
    runroot_path = runroot_tmp_path(runroot_key)
    mock_cfg_path = os.path.join(runroot_path, "mock.cfg")
    cmd = ["mock", "-r", mock_cfg_path] + args
    if log_output:
        stdout_log_path = os.path.join(runroot_path, "mock-stdout.log")
        stderr_log_path = os.path.join(runroot_path, "mock-stderr.log")
        with open(stdout_log_path, "a") as stdout_log:
            with open(stderr_log_path, "a") as stderr_log:
                execute_cmd(cmd, stdout=stdout_log, stderr=stderr_log)
    else:
        execute_cmd(cmd)


def mock_runroot_init(tag_name):
    """
    Creates and initializes new Mock chroot for runroot task.
    Prints the unique ID of chroot ("runroot_key") to stdout.

    :param str tag_name: Koji tag name from which the default packages for
        the Mock chroot are taken.
    """
    # Generate the runroot_key.
    runroot_key = str(uuid.uuid1())

    # Disable logging completely for the rest of execution, because the only
    # thing we must print is the runroot_key and the general logging to stdout
    # would confuse Pungi when calling this method.
    logging.disable(logging.CRITICAL)

    # At first run the cleanup task to remove possible old lost runroot
    # chroots.
    cleanup_old_runroots()

    # Get the latest Koji repo associated with the tag.
    koji_module = koji.get_profile_module(conf.koji_profile)
    koji_session = create_koji_session()
    repo = koji_session.getRepo(tag_name)
    if not repo:
        raise ValueError("Repository for tag %s does not exist." % tag_name)

    # Set the default options for Mock configuration.
    opts = {}
    opts["topdir"] = koji_module.pathinfo.topdir
    opts["topurl"] = koji_module.config.topurl
    opts["use_host_resolv"] = True
    opts["package_manager"] = "dnf"
    arch = koji.canonArch(platform.machine())

    # Generate the Mock configuration using the standard Koji way.
    output = koji_module.genMockConfig(
        runroot_key, arch, repoid=repo["id"], tag_name=tag_name, **opts
    )

    # Write the Mock configuration to /tmp/`runroot_key`/mock.cfg.
    mock_cfg_path = os.path.join(runroot_tmp_path(runroot_key), "mock.cfg")
    with open(mock_cfg_path, "w") as mock_cfg:
        mock_cfg.write(output)

    # Print the runroot_key to stdout, so the caller can store it and use it
    # in the future calls.
    print(runroot_key)

    # Run the `mock --init` with some log files.
    try:
        execute_mock(runroot_key, ["--init"])
    except Exception:
        rootdir = "/var/lib/mock/%s/root" % runroot_key
        mounts = [conf.target_dir] + conf.runroot_extra_mounts
        rmtree_skip_mounts(rootdir, mounts)
        raise


def raise_if_runroot_key_invalid(runroot_key):
    """
    Raise an ValueError exception in case the `runroot_key` contains forbidden
    characters.
    """
    for c in runroot_key:
        if c != "-" and not c.isalnum():
            raise ValueError(
                "Unexpected character '%s' in the runroot key \"%s\"."
                % (c, runroot_key)
            )


def mock_runroot_install(runroot_key, packages):
    """
    Installs the `packages` in the Mock chroot defined by `runroot_key`.

    :param str runroot_key: Runroot key.
    :param list packages: List of packages to install.
    """
    raise_if_runroot_key_invalid(runroot_key)
    try:
        execute_mock(runroot_key, ["--install"] + packages)
    except Exception:
        rootdir = "/var/lib/mock/%s/root" % runroot_key
        mounts = [conf.target_dir] + conf.runroot_extra_mounts
        rmtree_skip_mounts(rootdir, mounts)
        raise


def mock_runroot_run(runroot_key, cmd):
    """
    Executes the `cmd` in the Mock chroot defined by `runroot_key`.

    :param str runroot_key: Runroot key.
    :param list cmd: Command to execute.
    """
    raise_if_runroot_key_invalid(runroot_key)
    rootdir = "/var/lib/mock/%s/root" % runroot_key
    mounts = [conf.target_dir] + conf.runroot_extra_mounts

    try:
        # Mount the conf.targetdir in the Mock chroot.
        do_mounts(rootdir, mounts)

        # Wrap the runroot command in /bin/sh, because that's how Koji does
        # that and we need to stay compatible with this way...
        sh_wrapper = ["/bin/sh", "-c", "{ %s; }" % (" ".join(cmd))]

        # Run the command in Mock chroot. We need to use the `--old-chroot`
        # here, otherwise Lorax fails.
        args = ["--old-chroot", "--chroot", "--"] + sh_wrapper
        execute_mock(runroot_key, args, False)

    except Exception:
        # In the case of Exception, always call rmtree, because Pungi will
        # not call the "rpm -qa" last command, so we would never remove
        # the chroot.
        undo_mounts(rootdir, mounts)
        rmtree_skip_mounts(rootdir, mounts)
        raise

    # In the end of run, umount the conf.targetdir.
    undo_mounts(rootdir, mounts)
    # TODO: Pungi so far does not indicate anyhow that the runroot root can
    # be removed. The last command is always "rpm -qa ...", so we use it
    # as a mark for now that runroot root can be removed.
    # We should try improving Pungi OpenSSH runroot method to remove this
    # workaround later.
    if cmd[0] == "rpm":
        rmtree_skip_mounts(rootdir, mounts)


def mock_runroot_main(argv=None):
    """
    Main method handling the subcommands.

    :param list argv: List of arguments. If None, sys.argv is used.
    """
    argv = argv or sys.argv
    if argv[1] == "init":
        mock_runroot_init(argv[2])
    elif argv[1] == "install":
        mock_runroot_install(argv[2], argv[3:])
    elif argv[1] == "run":
        mock_runroot_run(argv[2], argv[3:])



"""Add index to Compose.state

Revision ID: 0571a5ca58a0
Revises: 0d4d8e1cfe29
Create Date: 2017-09-21 10:29:48.025577

"""

# revision identifiers, used by Alembic.
revision = "0571a5ca58a0"
down_revision = "0d4d8e1cfe29"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(op.f("ix_composes_state"), "composes", ["state"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_composes_state"), table_name="composes")
    # ### end Alembic commands ###



"""Add index to Compose.time_to_expire

Revision ID: 3b92820da295
Revises: 0571a5ca58a0
Create Date: 2017-09-21 11:03:04.006824

"""

# revision identifiers, used by Alembic.
revision = "3b92820da295"
down_revision = "0571a5ca58a0"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(
        op.f("ix_composes_time_to_expire"), "composes", ["time_to_expire"], unique=False
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_composes_time_to_expire"), table_name="composes")
    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>


import flask
import unittest

from mock import patch, Mock

import odcs.server.auth

from werkzeug.exceptions import Unauthorized
from odcs.server.auth import init_auth
from odcs.server.auth import load_krb_user_from_request
from odcs.server.auth import load_openidc_user
from odcs.server.auth import query_ldap_groups
from odcs.server.auth import require_scopes
from odcs.server.auth import load_krb_or_ssl_user_from_request
from odcs.server.auth import load_ssl_user_from_request
from odcs.server.errors import Forbidden
from odcs.server import app, conf, db
from odcs.server.models import User
from .utils import ModelsBaseTest


class TestLoadSSLUserFromRequest(ModelsBaseTest):
    def setUp(self):
        super(TestLoadSSLUserFromRequest, self).setUp()

        self.user = User(username="CN=tester1,L=prod,DC=example,DC=com")
        db.session.add(self.user)
        db.session.commit()

    def test_create_new_user(self):
        environ_base = {
            "SSL_CLIENT_VERIFY": "SUCCESS",
            "SSL_CLIENT_S_DN": "CN=client,L=prod,DC=example,DC=com",
        }

        with app.test_request_context(environ_base=environ_base):
            load_ssl_user_from_request(flask.request)

            expected_user = db.session.query(User).filter(
                User.username == "CN=client,L=prod,DC=example,DC=com"
            )[0]

            self.assertEqual(expected_user.id, flask.g.user.id)
            self.assertEqual(expected_user.username, flask.g.user.username)

            # Ensure user's groups are set to empty list
            self.assertEqual(0, len(flask.g.groups))

    def test_return_existing_user(self):
        environ_base = {
            "SSL_CLIENT_VERIFY": "SUCCESS",
            "SSL_CLIENT_S_DN": self.user.username,
        }

        with app.test_request_context(environ_base=environ_base):
            load_ssl_user_from_request(flask.request)

            self.assertEqual(self.user.id, flask.g.user.id)
            self.assertEqual(self.user.username, flask.g.user.username)

            # Ensure user's groups are set to empty list
            self.assertEqual(0, len(flask.g.groups))

    def test_401_if_ssl_client_verify_not_success(self):
        environ_base = {
            "SSL_CLIENT_VERIFY": "GENEROUS",
            "SSL_CLIENT_S_DN": self.user.username,
        }

        with app.test_request_context(environ_base=environ_base):
            with self.assertRaises(Unauthorized) as ctx:
                load_ssl_user_from_request(flask.request)
            self.assertIn("Cannot verify client: GENEROUS", ctx.exception.description)

    def test_401_if_cn_not_set(self):
        environ_base = {
            "SSL_CLIENT_VERIFY": "SUCCESS",
        }

        with app.test_request_context(environ_base=environ_base):
            with self.assertRaises(Unauthorized) as ctx:
                load_ssl_user_from_request(flask.request)
            self.assertIn(
                "Unable to get user information (DN) from client certificate",
                ctx.exception.description,
            )


class TestLoadKrbOrSSLUserFromRequest(unittest.TestCase):
    @patch("odcs.server.auth.load_ssl_user_from_request")
    @patch("odcs.server.auth.load_krb_user_from_request")
    def test_load_krb_or_ssl_user_from_request_remote_user(
        self, load_krb_user, load_ssl_user
    ):
        load_krb_user.return_value = "krb_user"
        load_ssl_user.return_value = "ssl_user"

        environ_base = {"REMOTE_USER": "newuser@EXAMPLE.COM"}

        with app.test_request_context(environ_base=environ_base):
            user = load_krb_or_ssl_user_from_request(flask.request)
            self.assertEqual(user, "krb_user")

    @patch("odcs.server.auth.load_ssl_user_from_request")
    @patch("odcs.server.auth.load_krb_user_from_request")
    def test_load_krb_or_ssl_user_from_request_ssl_client(
        self, load_krb_user, load_ssl_user
    ):
        load_krb_user.return_value = "krb_user"
        load_ssl_user.return_value = "ssl_user"

        environ_base = {
            "SSL_CLIENT_VERIFY": "SUCCESS",
            "SSL_CLIENT_S_DN": "ssl_user",
        }

        with app.test_request_context(environ_base=environ_base):
            user = load_krb_or_ssl_user_from_request(flask.request)
            self.assertEqual(user, "ssl_user")


class TestLoadKrbUserFromRequest(ModelsBaseTest):
    def setUp(self):
        super(TestLoadKrbUserFromRequest, self).setUp()

        self.user = User(username="tester1")
        db.session.add(self.user)
        db.session.commit()

    @patch("odcs.server.auth.query_ldap_groups")
    def test_create_new_user(self, query_ldap_groups):
        query_ldap_groups.return_value = ["devel", "admins"]

        environ_base = {"REMOTE_USER": "newuser@EXAMPLE.COM"}

        with app.test_request_context(environ_base=environ_base):
            load_krb_user_from_request(flask.request)

            expected_user = db.session.query(User).filter(User.username == "newuser")[0]

            self.assertEqual(expected_user.id, flask.g.user.id)
            self.assertEqual(expected_user.username, flask.g.user.username)

            # Ensure user's groups are created
            self.assertEqual(2, len(flask.g.groups))
            self.assertEqual(["admins", "devel"], sorted(flask.g.groups))

    @patch("odcs.server.auth.query_ldap_groups")
    def test_return_existing_user(self, query_ldap_groups):
        query_ldap_groups.return_value = ["devel", "admins"]
        original_users_count = db.session.query(User.id).count()

        environ_base = {"REMOTE_USER": "{0}@EXAMPLE.COM".format(self.user.username)}

        with app.test_request_context(environ_base=environ_base):
            load_krb_user_from_request(flask.request)

            self.assertEqual(original_users_count, db.session.query(User.id).count())
            self.assertEqual(self.user.id, flask.g.user.id)
            self.assertEqual(self.user.username, flask.g.user.username)
            self.assertEqual(["admins", "devel"], sorted(flask.g.groups))

    def test_401_if_remote_user_not_present(self):
        with app.test_request_context():
            with self.assertRaises(Unauthorized) as ctx:
                load_krb_user_from_request(flask.request)
            self.assertIn(
                "REMOTE_USER is not present in request.", ctx.exception.description
            )


class TestLoadOpenIDCUserFromRequest(ModelsBaseTest):
    def setUp(self):
        super(TestLoadOpenIDCUserFromRequest, self).setUp()

        self.user = User(username="tester1")
        db.session.add(self.user)
        db.session.commit()

    @patch("odcs.server.auth.requests.get")
    def test_create_new_user(self, get):
        get.return_value.status_code = 200
        get.return_value.json.return_value = {
            "groups": ["tester", "admin"],
            "name": "new_user",
        }

        environ_base = {
            "REMOTE_USER": "new_user",
            "OIDC_access_token": "39283",
            "OIDC_CLAIM_iss": "https://iddev.fedorainfracloud.org/openidc/",
            "OIDC_CLAIM_scope": "openid https://id.fedoraproject.org/scope/groups "
            "https://pagure.io/odcs/new-compose "
            "https://pagure.io/odcs/renew-compose "
            "https://pagure.io/odcs/delete-compose",
        }

        with app.test_request_context(environ_base=environ_base):
            load_openidc_user(flask.request)

            new_user = db.session.query(User).filter(User.username == "new_user")[0]

            self.assertEqual(new_user, flask.g.user)
            self.assertEqual("new_user", flask.g.user.username)
            self.assertEqual(sorted(["admin", "tester"]), sorted(flask.g.groups))

    @patch("odcs.server.auth.requests.get")
    def test_return_existing_user(self, get):
        get.return_value.status_code = 200
        get.return_value.json.return_value = {
            "groups": ["testers", "admins"],
            "name": self.user.username,
        }

        environ_base = {
            "REMOTE_USER": self.user.username,
            "OIDC_access_token": "39283",
            "OIDC_CLAIM_iss": "https://iddev.fedorainfracloud.org/openidc/",
            "OIDC_CLAIM_scope": "openid https://id.fedoraproject.org/scope/groups "
            "https://pagure.io/odcs/new-compose "
            "https://pagure.io/odcs/renew-compose "
            "https://pagure.io/odcs/delete-compose",
        }

        with app.test_request_context(environ_base=environ_base):
            original_users_count = db.session.query(User.id).count()

            load_openidc_user(flask.request)

            users_count = db.session.query(User.id).count()
            self.assertEqual(original_users_count, users_count)

            # Ensure existing user is set in g
            self.assertEqual(self.user.id, flask.g.user.id)
            self.assertEqual(["admins", "testers"], sorted(flask.g.groups))

    @patch("odcs.server.auth.requests.get")
    def test_user_info_failure(self, get):
        # If the user_info endpoint errors out, we continue to authenticate
        # based only on the user (which we have from the token), ignoring groups.
        get.return_value.status_code = 400

        environ_base = {
            "REMOTE_USER": self.user.username,
            "OIDC_access_token": "39283",
            "OIDC_CLAIM_iss": "https://iddev.fedorainfracloud.org/openidc/",
            "OIDC_CLAIM_scope": "openid https://id.fedoraproject.org/scope/groups "
            "https://pagure.io/odcs/new-compose "
            "https://pagure.io/odcs/renew-compose "
            "https://pagure.io/odcs/delete-compose",
        }

        with app.test_request_context(environ_base=environ_base):
            load_openidc_user(flask.request)

            self.assertEqual(self.user.id, flask.g.user.id)
            self.assertEqual([], sorted(flask.g.groups))

    def test_401_if_remote_user_not_present(self):
        environ_base = {
            # Missing REMOTE_USER here
            "OIDC_access_token": "39283",
            "OIDC_CLAIM_iss": "https://iddev.fedorainfracloud.org/openidc/",
            "OIDC_CLAIM_scope": "openid https://id.fedoraproject.org/scope/groups",
        }
        with app.test_request_context(environ_base=environ_base):
            self.assertRaises(Unauthorized, load_openidc_user, flask.request)

    def test_401_if_access_token_not_present(self):
        environ_base = {
            "REMOTE_USER": "tester1",
            # Missing OIDC_access_token here
            "OIDC_CLAIM_iss": "https://iddev.fedorainfracloud.org/openidc/",
            "OIDC_CLAIM_scope": "openid https://id.fedoraproject.org/scope/groups",
        }
        with app.test_request_context(environ_base=environ_base):
            self.assertRaises(Unauthorized, load_openidc_user, flask.request)

    def test_401_if_scope_not_present(self):
        environ_base = {
            "REMOTE_USER": "tester1",
            "OIDC_access_token": "39283",
            "OIDC_CLAIM_iss": "https://iddev.fedorainfracloud.org/openidc/",
            # Missing OIDC_CLAIM_scope here
        }
        with app.test_request_context(environ_base=environ_base):
            self.assertRaises(Unauthorized, load_openidc_user, flask.request)

    def test_401_if_required_scope_not_present_in_token_scope(self):
        environ_base = {
            "REMOTE_USER": "new_user",
            "OIDC_access_token": "39283",
            "OIDC_CLAIM_iss": "https://iddev.fedorainfracloud.org/openidc/",
            "OIDC_CLAIM_scope": "openid https://id.fedoraproject.org/scope/groups",
        }

        with patch.object(
            odcs.server.auth.conf, "auth_openidc_required_scopes", ["new-compose"]
        ):
            with app.test_request_context(environ_base=environ_base):
                with self.assertRaises(Unauthorized) as ctx:
                    load_openidc_user(flask.request)
                self.assertIn(
                    "Required OIDC scope new-compose not present.",
                    ctx.exception.description,
                )


class TestQueryLdapGroups(unittest.TestCase):
    """Test auth.query_ldap_groups"""

    @patch("odcs.server.auth.ldap.initialize")
    def test_get_groups(self, initialize):
        initialize.return_value.search_s.return_value = [
            (
                "cn=odcsdev,ou=Groups,dc=example,dc=com",
                {"gidNumber": ["5523"], "cn": ["odcsdev"]},
            ),
            (
                "cn=freshmakerdev,ou=Groups,dc=example,dc=com",
                {"gidNumber": ["17861"], "cn": ["freshmakerdev"]},
            ),
            (
                "cn=devel,ou=Groups,dc=example,dc=com",
                {"gidNumber": ["5781"], "cn": ["devel"]},
            ),
        ]

        groups = query_ldap_groups("me")
        self.assertEqual(sorted(["odcsdev", "freshmakerdev", "devel"]), sorted(groups))


class TestInitAuth(unittest.TestCase):
    """Test init_auth"""

    def setUp(self):
        self.login_manager = Mock()

    def test_select_kerberos_auth_backend(self):
        init_auth(self.login_manager, "kerberos")
        self.login_manager.request_loader.assert_called_once_with(
            load_krb_user_from_request
        )

    def test_select_openidc_auth_backend(self):
        init_auth(self.login_manager, "openidc")
        self.login_manager.request_loader.assert_called_once_with(load_openidc_user)

    def test_select_ssl_auth_backend(self):
        init_auth(self.login_manager, "ssl")
        self.login_manager.request_loader.assert_called_once_with(
            load_ssl_user_from_request
        )

    def test_select_kerberos_or_ssl_auth_backend(self):
        init_auth(self.login_manager, "kerberos_or_ssl")
        self.login_manager.request_loader.assert_called_once_with(
            load_krb_or_ssl_user_from_request
        )

    def test_not_use_auth_backend(self):
        init_auth(self.login_manager, "noauth")
        self.login_manager.request_loader.assert_not_called()

    def test_error_if_select_an_unknown_backend(self):
        self.assertRaises(ValueError, init_auth, self.login_manager, "xxx")
        self.assertRaises(ValueError, init_auth, self.login_manager, "")
        self.assertRaises(ValueError, init_auth, self.login_manager, None)

    def test_init_auth_no_ldap_server(self):
        with patch.object(odcs.server.auth.conf, "auth_ldap_server", ""):
            self.assertRaises(ValueError, init_auth, self.login_manager, "kerberos")

    def test_init_auths_no_ldap_group_base(self):
        with patch.object(odcs.server.auth.conf, "auth_ldap_group_base", ""):
            self.assertRaises(ValueError, init_auth, self.login_manager, "kerberos")


class TestDecoratorRequireScopes(unittest.TestCase):
    """Test decorator require_scopes"""

    @patch.object(conf, "oidc_base_namespace", new="http://example.com/")
    @patch.object(conf, "auth_backend", new="openidc")
    def test_function_is_called(self):
        with app.test_request_context():
            flask.g.oidc_scopes = ["http://example.com/renew-compose"]

            mock_func = Mock()
            mock_func.__name__ = "real_function"
            decorated_func = require_scopes("renew-compose")(mock_func)
            decorated_func(1, 2, 3)

        mock_func.assert_called_once_with(1, 2, 3)

    @patch.object(conf, "oidc_base_namespace", new="http://example.com/")
    @patch.object(conf, "auth_backend", new="openidc")
    def test_function_is_not_called_if_scope_is_not_present(self):
        with app.test_request_context():
            flask.g.oidc_scopes = [
                "http://example.com/new-compose",
                "http://example.com/renew-compose",
            ]

            mock_func = Mock()
            mock_func.__name__ = "real_function"
            decorated_func = require_scopes("delete-compose")(mock_func)
            self.assertRaises(Forbidden, decorated_func, 1, 2, 3)

    @patch.object(conf, "oidc_base_namespace", new="http://example.com/")
    @patch.object(conf, "auth_backend", new="kerberos")
    def test_function_is_called_for_non_openidc_backend(self):
        with app.test_request_context():
            flask.g.oidc_scopes = [
                "http://example.com/new-compose",
                "http://example.com/renew-compose",
            ]

            mock_func = Mock()
            mock_func.__name__ = "real_function"
            decorated_func = require_scopes("delete-compose")(mock_func)
            decorated_func(1, 2, 3)
            mock_func.assert_called_once_with(1, 2, 3)



"""Add Compose.removed_by

Revision ID: f4bc999818d7
Revises: 11b350234051
Create Date: 2018-02-23 17:17:46.664813

"""

# revision identifiers, used by Alembic.
revision = "f4bc999818d7"
down_revision = "11b350234051"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("composes", sa.Column("removed_by", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("composes", "removed_by")
    # ### end Alembic commands ###



"""Add modular_koji_tags and module_defaults_url columns.

Revision ID: e186faabdafe
Revises: b2725d046624
Create Date: 2019-01-28 08:15:35.059106

"""

# revision identifiers, used by Alembic.
revision = "e186faabdafe"
down_revision = "b2725d046624"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "composes", sa.Column("modular_koji_tags", sa.String(), nullable=True)
    )
    op.add_column(
        "composes", sa.Column("module_defaults_url", sa.String(), nullable=True)
    )


def downgrade():
    op.drop_column("composes", "module_defaults_url")
    op.drop_column("composes", "modular_koji_tags")



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import errno
import functools
import os
import signal
import time
import subprocess
import shutil
from threading import Timer
from six import text_type

from odcs.server import conf, log


def to_text_type(s):
    """
    Converts `s` to `text_type`. In case it fails, returns `s`.
    """
    try:
        return text_type(s, "utf-8")
    except TypeError:
        return s


def retry(
    timeout=conf.net_timeout,
    interval=conf.net_retry_interval,
    wait_on=Exception,
    logger=None,
):
    """A decorator that allows to retry a section of code until success or timeout."""

    def wrapper(function):
        @functools.wraps(function)
        def inner(*args, **kwargs):
            start = time.time()
            while True:
                try:
                    return function(*args, **kwargs)
                except wait_on as e:
                    if logger is not None:
                        logger.warn(
                            "Exception %r raised from %r.  Retry in %rs",
                            e,
                            function,
                            interval,
                        )
                    time.sleep(interval)
                    if (time.time() - start) >= timeout:
                        raise  # This re-raises the last exception.

        return inner

    return wrapper


def makedirs(path, mode=0o775):
    try:
        os.makedirs(path, mode=mode)
    except OSError as ex:
        if ex.errno != errno.EEXIST:
            raise


def _kill_process_group(proc, args):
    log.error("Timeout occured while running: %s", args)
    pgrp = os.getpgid(proc.pid)
    os.killpg(pgrp, signal.SIGINT)


def execute_cmd(args, stdout=None, stderr=None, cwd=None, timeout=None):
    """
    Executes command defined by `args`. If `stdout` or `stderr` is set to
    Python file object, the stderr/stdout output is redirecter to that file.
    If `cwd` is set, current working directory is set accordingly for the
    executed command.

    :param args: list defining the command to execute.
    :param stdout: Python file object to redirect the stdout to.
    :param stderr: Python file object to redirect the stderr to.
    :param cwd: string defining the current working directory for command.
    :param timeout: Timeout in seconds after which the process and all its
        children are killed.
    :raises RuntimeError: Raised when command exits with non-zero exit code.
    """
    out_log_msg = ""
    if stdout:
        out_log_msg += ", stdout log: %s" % stdout.name
    if stderr:
        out_log_msg += ", stderr log: %s" % stderr.name

    # Execute command and use `os.setsid` in preexec_fn to create new process
    # group so we can kill the main process and also children processes in
    # case of timeout.
    log.info("Executing command: %s%s" % (args, out_log_msg))
    proc = subprocess.Popen(
        args, stdout=stdout, stderr=stderr, cwd=cwd, preexec_fn=os.setsid
    )

    # Setup timer to kill whole process group if needed.
    if timeout:
        timeout_timer = Timer(timeout, _kill_process_group, [proc, args])

    try:
        if timeout:
            timeout_timer.start()
        proc.communicate()
    finally:
        timeout_expired = False
        if timeout:
            if timeout_timer.finished.is_set():
                timeout_expired = True
            timeout_timer.cancel()

    if timeout_expired:
        raise RuntimeError(
            "Compose has taken more time than allowed by configuration "
            "(%d seconds)" % conf.pungi_timeout
        )

    if proc.returncode != 0:
        err_msg = "Command '%s' returned non-zero value %d%s" % (
            args,
            proc.returncode,
            out_log_msg,
        )
        raise RuntimeError(err_msg)


def clone_repo(url, dest, branch="master", commit=None):
    cmd = ["git", "clone", "-b", branch, url, dest]
    execute_cmd(cmd)

    if commit:
        cmd = ["git", "checkout", commit]
        execute_cmd(cmd, cwd=dest)

    return dest


def copytree(src, dst, symlinks=False, ignore=None):
    """
    Implementation of shutil.copytree which does not fail when `dst` exists.
    """
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

COMPS_TEMPLATE = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE comps PUBLIC "-//Red Hat, Inc.//DTD Comps info//EN" "comps.dtd">
<comps>
  {%- for group in comps.groups %}
  <group>
    <id>{{ group.id }}</id>
    <name>{{ group.name }}</name>
    <description>{{ group.description }}</description>
    <default>{{ group.is_default|lower }}</default>
    <uservisible>{{ group.is_uservisible|lower }}</uservisible>
    <packagelist>
        {%- for package in group.packages %}
        <packagereq
        {%- if package.type %} type="{{ package.type }}"{% endif -%}
        {%- if package.arch %} arch="{{ package.arch }}"{% endif -%}
        {%- if package.requires %} requires="{{ package.requires }}"{% endif -%}
        {%- if package.is_basearchonly %} basearchonly="{{ package.is_basearchonly | lower }}"{% endif -%}
        {%- if false %}{% endif -%}>{{ package.name }}</packagereq>
        {%- endfor %}
    </packagelist>
  </group>
  {%- endfor %}
</comps>

"""

VARIANTS_TEMPLATE = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE variants PUBLIC "-//Red Hat, Inc.//DTD Variants info//EN" "variants2012.dtd">
<variants>
  {%- for variant in product.variants %}
  <variant id="{{ variant.id }}" name="{{ variant.name }}" type="{{ variant.type }}">
    <arches>
      {%- for arch in variant.arches %}
      <arch>{{ arch.name }}</arch>
      {%- endfor %}
    </arches>
    {%- if variant.groups %}
    <groups>
      {%- for group in variant.groups %}
        <group
        {%- if group in variant.default_groups %} default="true"{% endif -%}
        {%- if false %}{% endif %}>{{ group.name }}</group>
      {%- endfor %}
    </groups>
    {%- endif %}
    {%- if variant.modules %}
    <modules>
      {%- for module in variant.modules %}
      <module>{{ module.name }}</module>
      {%- endfor %}
    </modules>
    {%- endif %}
  </variant>
  {%- endfor %}
</variants>

"""


class Arch(object):
    def __init__(self, name):
        self.name = name


class Package(object):
    def __init__(
        self, name, arch=None, type=None, requires=None, is_basearchonly=False
    ):
        self.name = name
        self.arch = arch
        self.type = type
        self.requires = requires
        self.is_basearchonly = is_basearchonly


class Module(object):
    def __init__(self, name):
        self.name = name


class Group(object):
    def __init__(self, id, name, description, is_default=True, is_uservisible=True):
        self.id = id
        self.name = name
        self.description = description
        self.is_default = is_default
        self.is_uservisible = is_uservisible
        self.packages = []

    def add_package(self, package):
        self.packages.append(package)


class Comps(object):
    def __init__(self):
        self.groups = []

    def add_group(self, group):
        self.groups.append(group)


class Variant(object):
    def __init__(self, id, name, type, source_type):
        self.id = id
        self.name = name
        self.type = type
        self.arches = []
        self.source_type = source_type
        self.groups = []
        self.default_groups = []
        self.modules = []

    def add_arch(self, arch):
        self.arches.append(arch)

    def add_group(self, group, default=True):
        self.groups.append(group)
        if default:
            self.default_groups.append(group)

    def add_module(self, module):
        self.modules.append(module)


class Product(object):
    def __init__(self):
        self.variants = []

    def add_variant(self, variant):
        self.variants.append(variant)



"""empty message

Revision ID: de0a86d7de49
Revises: e186faabdafe
Create Date: 2019-09-24 12:31:18.648835

"""

# revision identifiers, used by Alembic.
revision = "de0a86d7de49"
down_revision = "e186faabdafe"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("composes", sa.Column("time_started", sa.DateTime(), nullable=True))
    # ### end Alembic commands ###
    # Set the start time for all composes that are not waiting.
    op.execute(
        "UPDATE composes SET time_started = time_submitted WHERE time_started IS NULL AND state != 0"
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("composes", "time_started")
    # ### end Alembic commands ###



import errno
from os import path, makedirs


# FIXME: workaround for this moment till confdir, dbdir (installdir etc.) are
# declared properly somewhere/somehow
confdir = path.abspath(path.dirname(__file__))
# use parent dir as dbdir else fallback to current dir
dbdir = (
    path.abspath(path.join(confdir, "../..")) if confdir.endswith("conf") else confdir
)


class BaseConfiguration(object):
    # Make this random (used to generate session keys)
    SECRET_KEY = "74d9e9f9cd40e66fc6c4c2e9987dce48df3ce98542529fd0"
    SQLALCHEMY_DATABASE_URI = "sqlite:///{0}".format(path.join(dbdir, "odcs.db"))
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    HOST = "127.0.0.1"
    PORT = 5005

    DEBUG = False
    # Global network-related values, in seconds
    NET_TIMEOUT = 120
    NET_RETRY_INTERVAL = 30

    # Available backends are: console, file, journal.
    LOG_BACKEND = "journal"

    # Path to log file when LOG_BACKEND is set to "file".
    LOG_FILE = "odcs.log"

    # Available log levels are: debug, info, warn, error.
    LOG_LEVEL = "info"

    SSL_ENABLED = False

    # Users are required to be in allowed_clients to generate composes,
    # you can add group names or usernames (it can be normal user or host
    # principal) into ALLOWED_CLIENTS. The group names are from ldap for
    # kerberos users or FAS for openidc users.
    #
    # You can also specify granular per-user or per-group permissions like
    # this:
    #
    # ALLOWED_CLIENTS = {
    #   'users': {
    #     'joe': {}, # Can use any flags/source_types/... as globally allowed
    #     'mike': {  # Can do only tag/module composes
    #       'source_types': ['tag', 'module'],
    #     },
    #     'ralph': { # Can use only 'no_deps' flags and 'repository' results
    #       'flags': ['no_deps'],
    #       'results': ['repository'],
    #     },
    #     'foo': { # Allow any source
    #       'source': [''],
    #     },
    #   }
    # }
    ALLOWED_CLIENTS = {
        "groups": {},
        "users": {},
    }

    # Users in ADMINS are granted with admin permission.
    ADMINS = {
        "groups": [],
        "users": [],
    }

    # OIDC base namespace
    # See also section pagure.io/odcs in
    # https://fedoraproject.org/wiki/Infrastructure/Authentication
    OIDC_BASE_NAMESPACE = "https://pagure.io/odcs/"

    # Select which authentication backend to work with. There are 3 choices
    # noauth: no authentication is enabled. Useful for development particularly.
    # kerberos: Kerberos authentication is enabled.
    # openidc: OpenIDC authentication is enabled.
    AUTH_BACKEND = ""

    # Used for Kerberos authentication and to query user's groups.
    # Format: ldap://hostname[:port]
    # For example: ldap://ldap.example.com/
    AUTH_LDAP_SERVER = ""

    # Group base to query groups from LDAP server.
    # Generally, it would be, for example, ou=groups,dc=example,dc=com
    AUTH_LDAP_GROUP_BASE = ""

    AUTH_OPENIDC_USERINFO_URI = "https://id.fedoraproject.org/openidc/UserInfo"

    # Scope requested from Fedora Infra for permission of submitting request to
    # run a new compose.
    # See also: https://fedoraproject.org/wiki/Infrastructure/Authentication
    # Add additional required scope in following list
    #
    # ODCS has additional scopes, which will be checked later when specific
    # API is called.
    # https://pagure.io/odcs/new-compose
    # https://pagure.io/odcs/renew-compose
    # https://pagure.io/odcs/delete-compose
    AUTH_OPENIDC_REQUIRED_SCOPES = [
        "openid",
        "https://id.fedoraproject.org/scope/groups",
    ]

    # Select backend where message will be sent to. Currently, umb is supported
    # which means the Unified Message Bus.
    MESSAGING_BACKEND = ""  # fedora-messaging or umb

    # List of broker URLs. Each of them is a string consisting of domain and
    # optiona port.
    MESSAGING_BROKER_URLS = []

    # Path to certificate file used to authenticate ODCS by messaging broker.
    MESSAGING_CERT_FILE = ""

    # Path to private key file used to authenticate ODCS by messaging broker.
    MESSAGING_KEY_FILE = ""

    MESSAGING_CA_CERT = ""

    # The MESSAGING_TOPIC is used as topic for messages sent when compose
    # state is change.
    # The INTERNAL_MESSAGING_TOPIC is used for ODCS internal messages sent
    # from frontends to backends. It for example triggers removal of expired
    # composes.
    # For umb, it is the ActiveMQ virtual topic e.g.
    # VirtualTopic.eng.odcs.state.changed.
    MESSAGING_TOPIC = ""
    INTERNAL_MESSAGING_TOPIC = ""

    # Definitions of raw Pungi configs for "raw_config" source_type.
    # RAW_CONFIG_URLS = {
    #   "my_raw_config": {
    #       "url": "http://example.com/test.git",
    #       "config_filename": "pungi.conf",
    #       "path": "some/git/subpath",  # optional
    #   }
    # }

    # Command line arguments used to construct pungi-koji command.
    PUNGI_KOJI_ARGS = []

    # Command line argument for raw_config source type, which overwrite
    # arguments listed PUNGI_KOJI_ARGS.
    # If a particular raw config should have specific pungi-koji arguments, add
    # a key/value into this option, where key should exist in the
    # RAW_CONFIG_URLS, and value lists each arguments.
    # If no argument is required, just omit it, or add a key/value just as
    # mentioned, but keep value as a empty list.
    # RAW_CONFIG_PUNGI_KOJI_ARGS = {
    #     'my_raw_config': ['arg1', ...],
    #     'another_raw_config': ['arg1', ...],
    # }


class DevConfiguration(BaseConfiguration):
    DEBUG = True
    LOG_BACKEND = "console"
    LOG_LEVEL = "debug"

    # Global network-related values, in seconds
    NET_TIMEOUT = 5
    NET_RETRY_INTERVAL = 1
    TARGET_DIR = path.join(dbdir, "test_composes")
    try:
        makedirs(TARGET_DIR, mode=0o775)
    except OSError as ex:
        if ex.errno != errno.EEXIST:
            raise RuntimeError(
                "Can't create compose target dir %s: %s" % (TARGET_DIR, ex.strerror)
            )

    PUNGI_CONF_PATH = path.join(confdir, "pungi.conf")
    AUTH_BACKEND = "noauth"
    AUTH_OPENIDC_USERINFO_URI = "https://iddev.fedorainfracloud.org/openidc/UserInfo"

    KOJI_PROFILE = "stg"

    RAW_CONFIG_WRAPPER_CONF_PATH = path.join(confdir, "raw_config_wrapper.conf")


class TestConfiguration(BaseConfiguration):
    LOG_BACKEND = "console"
    LOG_LEVEL = "debug"
    DEBUG = True

    # Use in-memory sqlite db to make tests fast.
    SQLALCHEMY_DATABASE_URI = "sqlite://"

    PUNGI_CONF_PATH = path.join(confdir, "pungi.conf")
    RAW_CONFIG_WRAPPER_CONF_PATH = path.join(confdir, "raw_config_wrapper.conf")
    # Global network-related values, in seconds
    NET_TIMEOUT = 0
    NET_RETRY_INTERVAL = 0
    TARGET_DIR = path.join(dbdir, "test_composes")
    try:
        makedirs(TARGET_DIR, mode=0o775)
    except OSError as ex:
        if ex.errno != errno.EEXIST:
            raise RuntimeError(
                "Can't create compose target dir %s: %s" % (TARGET_DIR, ex.strerror)
            )

    AUTH_BACKEND = "noauth"
    AUTH_LDAP_SERVER = "ldap://ldap.example.com"
    AUTH_LDAP_GROUP_BASE = "ou=groups,dc=example,dc=com"

    MESSAGING_BACKEND = "rhmsg"
    KOJI_PROFILE = "koji"


class ProdConfiguration(BaseConfiguration):
    pass
